//////////////////////////////////////////////////////////////////////////
/// \file 			MFCCatalogExample.cpp 
/// \brief 			Small example showing catalog use.
/// \date 			2019 
/// \copyright 		Metrolab Technology SA
/// \author 		Simon Tellier
// 
//

#include <iostream>
#include <queue>
#include <fstream>		// ifstream to read files
#include <windows.h>	// GetModuleFileName

#include "InstrumentCatalog.h"
#include "MFCManager.h"

using namespace MTL;

class CMyApplication
{
private:
	CGUICatalogInterface					m_Cat;
	CAsyncSnaphsotNotifier					m_ASNotifier;
	std::mutex								m_QLock;
	// This queue is the heart of the interchange mechanism. 
	// The catalog being "interchange-agnostic", it is up to you to 
	// select which type of communication mechanism you want to implement 
	// in your application. In this specific and simple example, we
	// aimed for a queue.
	std::queue< CAsyncNotification *>		m_Q;

private:
	void l_PushNewAsyncNotification(CAsyncNotification * pANot)
	{
		std::lock_guard<std::mutex> l_LG(m_QLock);
		m_Q.push(pANot);
	}
	CAsyncNotification * l_GetNotification()
	{
		std::lock_guard<std::mutex> l_LG(m_QLock);
		if (m_Q.empty())
			return nullptr;

		CAsyncNotification * l_pNot = m_Q.front();
		m_Q.pop();
		return l_pNot;
	}
public:
	CMyApplication(CInstrumentCatalog & rCat)
		: m_Cat(rCat), m_ASNotifier(&CMyApplication::l_PushNewAsyncNotification, this)
	{
	}
	void SyncControls()
	{
		// We do not need to sync anything with the catalog for this application
	}
	void SyncObservers()
	{
		// We do not need to sync anything with the catalog for this application
	}
	void Subscribe()
	{
		// Proceed to the subscription to each variable of interrest.
		m_Cat.oInstrumentList.Subscribe(m_ASNotifier);
		m_Cat.oProbeArrayInformation.Subscribe(m_ASNotifier);
		m_Cat.oSearchResult.Subscribe(m_ASNotifier);
		m_Cat.oInstrumentStatus.Subscribe(m_ASNotifier);
		m_Cat.oError.Subscribe(m_ASNotifier);
	}

	void DoJob()
	{
		bool l_Operate = true;
		while (l_Operate)
		{
			// Get the current notification.
			CAsyncNotification * l_pNot = l_GetNotification();
			if (nullptr == l_pNot)
			{	
				// No notification currently in the queue. 
				// Sleep for a while.
				std::this_thread::sleep_for(std::chrono::milliseconds(100));
			}
			else
			{
				// Since we have a notification...
				CObserver_untyped & rVar = l_pNot->Observer();
				const CAsyncSnapshotNotification * l_pASNot = static_cast<CAsyncSnapshotNotification *>(l_pNot);

				// try determining which variable spawned it.
				if (rVar.Is(m_Cat.oInstrumentList))
				{
					std::cout << "New Instrument List" << std::endl;
					const CInstrumentCatalog::tInstrumentList * l_pVal = static_cast<const CInstrumentCatalog::tInstrumentList *>(l_pASNot->DataCopy());
					if (!l_pVal->empty())
						m_Cat.cInstrumentCurrent(l_pVal->at(0));
				}
				else if (rVar.Is(m_Cat.oProbeArrayInformation))
				{
					std::cout << "New Probe-Array Information" << std::endl;
					const CInstrumentCatalog::sProbeArrayInformation * l_pVal = static_cast<const CInstrumentCatalog::sProbeArrayInformation *>(l_pASNot->DataCopy());
					if (l_pVal->Type != CInstrumentCatalog::eProbeArrayType::kUnknown)
					{
						CInstrumentCatalog::sInstrumentState l_NewState;
						l_NewState.OperatingMode = CInstrumentCatalog::eOperatingMode::kSearch;
						l_NewState.RepeatMode = CInstrumentCatalog::eRepeatMode::kSingle;
						l_NewState.ProbeSelection = CInstrumentCatalog::sMeasureProbeSelection();
						l_NewState.FreqMin = l_pVal->FreqMin;
						l_NewState.FreqMax = l_pVal->FreqMax;
						l_NewState.NbAveragedMeasurements = 0;
						m_Cat.cInstrumentState(l_NewState);
					}
				}
				else if (rVar.Is(m_Cat.oSearchResult))
				{
					const CInstrumentCatalog::sSearchResult * l_pVal = static_cast<const CInstrumentCatalog::sSearchResult *>(l_pASNot->DataCopy());
					std::cout << "New Search Result :" << CInstrumentCatalog::l_Dump(*l_pVal);
					l_Operate = false;
				}
				else if (rVar.Is(m_Cat.oInstrumentStatus))
				{
					const CInstrumentCatalog::sInstrumentStatus * l_pVal = static_cast<const CInstrumentCatalog::sInstrumentStatus *>(l_pASNot->DataCopy());
					std::cout << CInstrumentCatalog::l_Dump(*l_pVal);
				}
				else if (rVar.Is(m_Cat.oError))
				{
					const CInstrumentCatalog::sError * l_pVal = static_cast<const CInstrumentCatalog::sError *>(l_pASNot->DataCopy());
					std::cout << CInstrumentCatalog::l_Dump(*l_pVal);
					l_Operate = false;
				}

				// Delete notification (and its content)
				delete l_pNot;
			}
		}
	}
};

std::string GetApplicationDir()
{
#ifdef _WIN32
	wchar_t l_ExePath[MAX_PATH];
	GetModuleFileName(NULL, l_ExePath, MAX_PATH);					// Get executable path
	std::wstring l_AppPath(l_ExePath);
	std::string l_AppDir(l_AppPath.begin(), l_AppPath.end());
	return l_AppDir.substr(0, l_AppDir.find_last_of('\\'));			// Remove application name
#endif
}

int main()
{
	std::string l_AppDir = GetApplicationDir();

	CInstrumentCatalog		l_Catalog;
	CMFCManager				l_MFCMan(l_Catalog, l_AppDir + "/../../../../InstrumentCatalog/MFC/ExternalFiles/DefaultCommandScript.scs", l_AppDir + "/../../../../InstrumentCatalog/MFC/ExternalFiles");
	CMyApplication			l_MyApp(l_Catalog);


	// Synchronize catalog controls
	l_MFCMan.SyncControls();
	l_MyApp.SyncControls();
	// Synchronize catalog observers
	l_MFCMan.SyncObservers();
	l_MyApp.SyncObservers();
	// Subscribe to variables
	l_MFCMan.Subscribe();
	l_MyApp.Subscribe();
	// Operate
	l_MFCMan.Operate();

	l_MyApp.DoJob();

	// Terminate manager operation
	l_MFCMan.Terminate();

	return 0;
}
