///	\file			CPT2026PeripheralROM.h
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#pragma once

// Standard includes
#include <assert.h>
#include <unordered_map>

// Personal includes
#include "PT2026Types.h"
#include "OSDefines.h"

using namespace MTL::Instrument;
using namespace MTL::Instrument::PT2026Types;

namespace MTL {

	class CPT2026PeripheralROM
	{
	//----------------------------------------------------------------------//
	//						Public Type Definitions							//
	//----------------------------------------------------------------------//
	public:
		enum eHeaderType { kUnknown, kStandalone, kEmbedded };

		//==================================================================================================//
		//											Common													//
		//==================================================================================================//
		enum eDeviceType {
			kUndefined = 0,
			kSingleContinuous = 1,
			kSinglePulsed = 2,
			kMultiplexer = 3,
			kSwitchedContinuous = 4,
			kSwitchedPulsed = 5,
			kCameraPulsed = 6
		};
		struct sIdentification
		{
			U16			ProtocolVersion;
			eDeviceType	DeviceType;
			U16			Model;
			U32			SN;		// Effective size is 3 bytes long
			U32			Capabilities;

			static const size_t SN_LEN = 3;
			static const size_t IDENTIFICATION_LEN = sizeof(U16) + sizeof(U8) + sizeof(U16) + SN_LEN + sizeof(U32);
		};
		struct sManufactureInformation {
			std::string ManufacturingDate;
			std::string Designation;
		};
		struct sStandaloneStaticHeader
		{
			static const U16 VERSION = 0x0001;
			static const size_t DATE_LEN = 10;
			static const size_t HEADER_LEN = sizeof(U16) + sizeof(U16) + sizeof(U16) + DATE_LEN + sIdentification::IDENTIFICATION_LEN + sizeof(U8);

			U16					ROMVersion;						// Version of this structure in BCD format xx.yy
			U16					ROMCapacity;					// Memory capacity in number of bytes
			U16					ROMSize;						// Usage of the above capacity in number of bytes
			char				ROMProgrammingDate[DATE_LEN];	// Format: ISO-8601 extended calendar date "YYYY-MM-DD"
			sIdentification		Identification;					// Identification of the device hosting the ROM
			U8					HeaderCRC;						// CRC of the header
		};
		struct sEmbeddedStaticHeader
		{
			static const U16 VERSION = 0x0002;
			static const size_t DATE_LEN = 10;
			static const size_t HEADER_LEN = sizeof(U16) + sizeof(U16) + sizeof(U16) + DATE_LEN + sizeof(U8);

			U16					ROMVersion;						// Version of this structure in BCD format xx.yy
			U16					ROMCapacity;					// Memory capacity in number of bytes
			U16					ROMSize;						// Usage of the above capacity in number of bytes
			char				ROMProgrammingDate[DATE_LEN];	// Format: ISO-8601 extended calendar date "YYYY-MM-DD"
			U8					HeaderCRC;						// CRC of the header
		};	
		class CParmInfoMap : public std::unordered_map<U8, U16>		// Composed of ParmNumber (U8) and its Address (U16)
		{
		public:
			static const size_t PARM_INFO_LEN = sizeof(U8) + sizeof(U16);
		};

		//==================================================================================================//
		//											Pulsed Wave												//
		//==================================================================================================//
		struct sPwPhysicalInformation
		{
			F64					MinFreq;					// Minimum probe frequency				[Hz]
			F64					MaxFreq;					// Maximum probe frequency				[Hz]
			F64					MinIF;						// Minimum IF frequency					[Hz]
			F64					MaxIF;						// Maximum IF frequency					[Hz]
			F64					FreqMultiplier;				// Frequency multiplication factor		[-]
			F64					GyroRatio;					// Gyromagnetic ratio					[Hz/T]
			F64					PulsePeriod;				// Usually spin relaxing time T1		[s]
			F64					T2;							// Spin relaxing time T2				[s]
			F64					SampleLength;				// Sample length						[s]
			U8					DeadTime;					// Dead time							[us]
			std::vector<U16>	NoiseLevelCurve;			// Noise level array					[mV]
			std::vector<U16>	TuningCurve;				// Tuning array							[mV]
			std::vector<U16>	MatchingCurve;				// Matching array						[mV]
			std::vector<U16>	PulseWidthCurve;			// Pulse width array					[us]
			std::vector<U16>	FreqStepCurve;				// Frequency steps array				[Hz]

			static const U32 MAX_CURVE_SIZE = 128;
			static const U16 HANDLED_PWP_PHYSICAL_INFORMATION_VERSION = 0x0001;
		};

		struct sPwPhysicalInformationSet
		{
		private:
			std::vector<sPwPhysicalInformation*>	pPhysInfoList;

		public:
			typedef std::vector<sPwPhysicalInformation*>::size_type size_type;

			virtual ~sPwPhysicalInformationSet(void)
			{
				for (std::vector<sPwPhysicalInformation*>::iterator it = pPhysInfoList.begin(); it != pPhysInfoList.end(); it++)
					(*it)->~sPwPhysicalInformation();
			}
			void Attach(sPwPhysicalInformation* pPwPhysInfo)
			{
				pPhysInfoList.push_back(pPwPhysInfo);
			}
			size_type size(void) const
			{
				return pPhysInfoList.size();
			}
			sPwPhysicalInformation& operator[] (const size_type Index)
			{
				assert(Index < pPhysInfoList.size());
				return *(pPhysInfoList[Index]);
			}
			sPwPhysicalInformation& operator[] (const size_type Index) const
			{
				assert(Index < pPhysInfoList.size());
				return *(pPhysInfoList[Index]);
			}
		};

		struct sTransmissionTimes
		{
			U32	NMRTransTime_us;
			U32	ExtNMRTransTime_us;
			sTransmissionTimes()
				: NMRTransTime_us(0), ExtNMRTransTime_us(0)
			{}
			sTransmissionTimes(U32 Short_us, U32 Extended_us)
				: NMRTransTime_us(Short_us), ExtNMRTransTime_us(Extended_us)
			{}
		};
		typedef std::vector<sTransmissionTimes>		tTransmissionTimeList;

		struct sPwFCAPhysInfo
		{
			F64		MinFreq;					// Minimum probe frequency				[Hz]
			F64		MaxFreq;					// Maximum probe frequency				[Hz]
			F64		MinIF;						// Minimum IF frequency					[Hz]
			F64		MaxIF;						// Maximum IF frequency					[Hz]
			F64		FreqMultiplier;				// Frequency multiplication factor		[-]
			U16		VoltageNoise;				// Noise level across the bandpass		[mV]

			sPwFCAPhysInfo(void)
				: MinFreq(0), MaxFreq(0), MinIF(0), MaxIF(0), FreqMultiplier(1), VoltageNoise(0)
			{}

			static const U16 HANDLED_PW_FCA_PHYSICAL_INFORMATION_VERSION = 0x0001;
		};

		//==================================================================================================//
		//											Multi Head												//
		//==================================================================================================//
		struct sHeadInfo
		{
			static const U16	SIZE = 3;

			U8	MUXcode;			// Multiplexing code
			U8	Info;				// Information bitfield (7-1: NA, 0 : IsEM8126)
			U8	PhysInfoSetIndex;	// Index of the physical information for this head
		};
		struct sMultiHeadInfo
		{
			static const U16	STRUCT_VERSION = 0x0002;

			U16		StructVersion;
			U16		OffsetDAC;			// DAC value for offset compensation
			std::vector<sHeadInfo>	MultiHeadInfo;
		};

		typedef std::vector<U8> tHeadList;

		struct sHeadDefinition
		{
			U8	HeadNumber;			// Head number among the multiple heads
			U8	PhysInfoSetIndex;	// Index of the physical information for this head

			sHeadDefinition(U8 HeadNum, U8 PhysInfInd)
				: HeadNumber(HeadNum), PhysInfoSetIndex(PhysInfInd)
			{}
		};
		#define HEAD_DEF_SIZE		2
		typedef std::vector<sHeadDefinition> tHeadDefList;

		//==================================================================================================//
		//											Camera													//
		//==================================================================================================//
		typedef std::vector<F32>	tNormalizationTable;	// [Hz]

		struct sPwCamPhysicalInfo {
			U16					Version;
			F64					LowFreq;				// [Hz]
			F64					NominalFreq;			// [Hz]
			F64					HighFreq;				// [Hz]
			F64					GyroRatio;				// [Hz/T]
			std::string			NormalizationDate;		// YYYY-MM-DD
			tNormalizationTable	NormalizationTable;

			static const U16 HANDLED_PW_CAM_PHYSICAL_INFORMATION_VERSION = 0x0002;
		};

		//==================================================================================================//
		//											Hall													//
		//==================================================================================================//
		struct sHallInformation {
			static const U16	STRUCT_VERSION = 0x0002;
			// State definition is as follows:
			// bits 7-2	: NA
			// bit 1	: Activated (Whether the hall sensor has to be used or not, for any reason)
			// bit 0	: Calibrated (Whether Calibration struct contains valid calibration data or not)
			U8		State;					// State of the hall sensor
			U16		Range_mT;				// Hall sensor range in [mT]
			U8		Tolerance;				// Calibration tolerance [%]
			struct sCalibration {
				// Coefficients correspond to a field B expressed in [T]
				U16		B0X, B0Y, B0Z;		// [bADC]	B field offset correction factor for each axis
				F32		G0X, G0Y, G0Z;		// [T/bADC]	B field gain correction factor for each axis
				U16		T0;					// [bADC]	Temperature offset correction factor, common to all axis
				F32		CT0X, CT0Y, CT0Z;	// [T/bADC]	Offset Temperature correction factor for each axis
				F32		CTGX, CTGY, CTGZ;	// [T/bADC]	Gain Temperature correction factor for each axis

				sCalibration()
					: B0X(32768), B0Y(32768), B0Z(32768), G0X(1), G0Y(1), G0Z(1), T0(0), CT0X(0), CT0Y(0), CT0Z(0), CTGX(0), CTGY(0), CTGZ(0)
				{}

				static F32 RawBtoB(const U16 & rRawB, const U16 & rRawT, const U16 & rB0, const F32 & rG0, const U16 & rT0, const F32 & rCT0, const F32 & rCTG)
				{
					I32 l_BOffsetless = (I32)rRawB - (I32)rB0;
					I32 l_TOffsetless = (I32)rRawT - (I32)rT0;
					F32 l_TOffsetCorrected = l_TOffsetless * rCT0;
					F32 l_TGainCorrected = l_TOffsetless * rCTG;
					F32 l_Gain = rG0 + l_TGainCorrected;
					F32 l_BCorrected_T = (l_BOffsetless - l_TOffsetCorrected) * l_Gain;
					return l_BCorrected_T;
				}

				static F32 Norm(const F32 & rBx, const F32 & rBy, const F32 & rBz)
				{
					return sqrt(rBx*rBx + rBy * rBy + rBz * rBz);
				}

				void Calibrate(const U16 & rRawBx, const U16 & rRawBy, const U16 & rRawBz, const U16 & rRawT,
					F32 & rBx_T, F32 & rBy_T, F32 & rBz_T, F32 & rNorm_T)
				{
					F32 Bx_T = RawBtoB(rRawBx, rRawT, B0X, G0X, T0, CT0X, CTGX);
					F32 By_T = RawBtoB(rRawBy, rRawT, B0Y, G0Y, T0, CT0Y, CTGY);
					F32 Bz_T = RawBtoB(rRawBz, rRawT, B0Z, G0Z, T0, CT0Z, CTGZ);
					rNorm_T = Norm(Bx_T, By_T, Bz_T);
					rBx_T = Bx_T;
					rBy_T = By_T;
					rBz_T = Bz_T;
				}
			} Calibration;

			bool IsCalibrated(void)
			{
				return State & 0x01;
			}
			bool IsActivated(void)
			{
				return State & 0x02;
			}
		};

		//==================================================================================================//
		//												Devices												//
		//==================================================================================================//
		enum class e1226ROMParm {
			kManufactureInformation		= 0,
			kPwPhysicalInformation		= 1,
			kHallInformation			= 2
		};
		enum class e9046ROMParm {
			kManufactureInformation = 0,
			kMultiHeadInformation = 1,
			kSearchHeads = 2,
			kMeasHeads = 3,
			kPwCamPhysicalInformation = 4,
			kPwPhysicalInformationSet = 5
		};



		//----------------------------------------------------------------------//
		//								Parser Class							//
		//----------------------------------------------------------------------//
		class Parser
		{
			//----------------------------------------------------------------------//
			//							Private Attributes							//
			//----------------------------------------------------------------------//
		private:
			const std::vector<U8>::const_iterator		m_beg, m_end;
			CParmInfoMap								m_ParmInfoMap;
			eHeaderType									m_HeaderType = kUnknown;
			sStandaloneStaticHeader						m_StdandaloneHeader;
			sEmbeddedStaticHeader						m_EmbeddedHeader;

			//----------------------------------------------------------------------//
			//							Private Methods								//
			//----------------------------------------------------------------------//
		private:
			bool l_ParseIdentification(sIdentification & rId, std::vector<U8>::const_iterator Id_beg, std::vector<U8>::const_iterator Id_end);
			bool l_HeaderVersion(U16 & rVersion);
			bool l_ParseStaticHeader(sStandaloneStaticHeader & rHeader);
			bool l_ParseStaticHeader(sEmbeddedStaticHeader & rHeader);
			bool l_ParseParmInfo(size_t HeaderLen, CParmInfoMap & rParmInfoMap);

			//----------------------------------------------------------------------//
			//							Public Methods								//
			//----------------------------------------------------------------------//
		public:
			Parser(std::vector<U8>::const_iterator beg, std::vector<U8>::const_iterator end)
				: m_beg(beg), m_end(end)
			{}

			bool Parse();

			bool HasIdentification(sIdentification & rID);
			bool ParmAddress(U8 ParmNo, U16 & rParmAddress);
			eHeaderType HeaderType();
			sStandaloneStaticHeader Header(const sStandaloneStaticHeader & rHeader);
			sEmbeddedStaticHeader Header(const sEmbeddedStaticHeader & rHeader);
			bool ParseManufactureInformation(size_t ParmAddress, sManufactureInformation & rManuInfo);
			bool ParseMultiHeadInfo(size_t ParmAddress, sMultiHeadInfo & rMultiHeadInfo);
			bool ParseHeadList(size_t ParmAddress, tHeadList & rHeadList);
			bool ParsePwCamPhysInfo(size_t ParmAddress, sPwCamPhysicalInfo & rPhysInfo);
			bool ParsePwPhysicalInfo(size_t ParmAddress, sPwPhysicalInformation & rPhysInfo, std::vector<U8>::const_iterator & rpBuffer);
			bool ParsePwPhysInfoSet(size_t ParmAddress, sPwPhysicalInformationSet & rPwPhysInfoSet);
			bool ParsePwFcaPhysInfo(size_t ParmAddress, sPwFCAPhysInfo & rPhysInfo);
			bool ParseHallInfo(size_t ParmAddress, sHallInformation & rHallInfo);
		};

		//----------------------------------------------------------------------//
		//								Builder Class							//
		//----------------------------------------------------------------------//
		class Builder
		{
		public:
			struct sParmContent
			{
				U8							ParmNumber;
				const std::vector<U8> &		rParmData;
			};
			typedef std::vector<sParmContent>	tParmContentInfo;

		private:
			static CParmInfoMap l_BuildParmInfo(const tParmContentInfo & rParmContent, U16 BaseAddress, size_t & rParmDataSize);
			static CParmInfoMap l_BuildParmInfo(const tParmContentInfo & rParmContent, U16 BaseAddress, std::vector<U8> & rGeneratedParmData);

			static std::vector<U8> l_Raw(const CPT2026PeripheralROM::sIdentification & rId);
			static std::vector<U8> l_Raw(const sStandaloneStaticHeader & rHeader);
			static std::vector<U8> l_Raw(const sEmbeddedStaticHeader & rHeader);
			static std::vector<U8> l_Raw(const CParmInfoMap & rParmInfo);

			template <typename HeaderType>
			static std::vector<U8> l_BuildROM(HeaderType Header, const tParmContentInfo & rParmContent);

		public:
			static std::vector<U8> Raw(const CPT2026PeripheralROM::sHallInformation & rHallInfo);
			static std::vector<U8> Raw(const CPT2026PeripheralROM::sManufactureInformation & rManInfo);
			static std::vector<U8> Raw(const CPT2026PeripheralROM::sMultiHeadInfo & rMultHeadInfo);
			static std::vector<U8> Raw(const CPT2026PeripheralROM::tHeadList & rHeadList);
			static std::vector<U8> Raw(const CPT2026PeripheralROM::sPwCamPhysicalInfo & rPwCamPhysInfo);
			static std::vector<U8> Raw(const CPT2026PeripheralROM::sPwFCAPhysInfo & rPwFCAPhysInfo);
			static std::vector<U8> Raw(const CPT2026PeripheralROM::sPwPhysicalInformation & rPwPhysInfo);
			static std::vector<U8> Raw(const CPT2026PeripheralROM::sPwPhysicalInformationSet & rPwPhysInfoSet);

			static std::vector<U8> BuildROM(sStandaloneStaticHeader Header, const tParmContentInfo & rParmContent);
			static std::vector<U8> BuildROM(sEmbeddedStaticHeader Header, const tParmContentInfo & rParmContent);

			static void Build1226ROM(	std::vector<U8> & rOutputROM,
										const CPT2026PeripheralROM::sManufactureInformation & rManInfo,
										const CPT2026PeripheralROM::sPwPhysicalInformation & rPwPhysInfo,
										const CPT2026PeripheralROM::sHallInformation & rHallInfo);
			static void Build9046ROM(	std::vector<U8> & rOutputROM,
										const CPT2026PeripheralROM::sIdentification & rIdentification,
										const CPT2026PeripheralROM::sManufactureInformation & rManInfo,
										const CPT2026PeripheralROM::sMultiHeadInfo & rMultiHeadInfo,
										const CPT2026PeripheralROM::tHeadList & rSearchHeadList,
										const CPT2026PeripheralROM::tHeadList & rMeasHeadList,
										const CPT2026PeripheralROM::sPwCamPhysicalInfo & rPwCamPhysInfo,
										const CPT2026PeripheralROM::sPwPhysicalInformationSet & rPwPhysInfoSet);
		};
	};

} // namespace MTL
