///	\file			InstrumentCatalog.h
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#pragma once

#include "CatalogInterface.h"
#include "OSDefines.h"

#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <ctime>

using namespace MTL;
using namespace MTL::InstrumentCatalogInterface;

namespace MTL {
	class CInstrumentCatalog {
	public:
		//----------------------------------------------------------------------//
		//								Types									//
		//----------------------------------------------------------------------//
		typedef std::time_t		tPosixTime;			// Unix Time / Posix Time / Epoch time e.g. defined as the number of seconds that have elapsed since 00:00:00 Coordinated Universal Time (UTC), Thursday, 1 January 1970
		typedef std::string		tISO8601DateTime;	// Date and Time in the ISO8601 YYYY-mm-ddTHH:MM:SSZ format (UTC Date and time)

		typedef std::string		tInstrumentLabel;		// Instrument name to be presented to the customer (also called "pretty name")
		typedef std::string		tInstrumentUniqueID;	// Unique instrument ID to identify the instrument (corresponds to the VISA resource name)
		enum class eConnectionInterface {			// Type of interface for a specific connection
			kUnknown,									// Unknown interface
			kUSB,										// USB interface
			kEthernet,									// Ethernet interface
			kSerial										// Serial interface
		};
		struct sInstrumentConnection {				// Describes an available connection to an instrument
			tInstrumentLabel		Label;				// Instrument label
			tInstrumentUniqueID		UniqueID;			// Insturment unique ID
			eConnectionInterface	Interface;			// Interface type
			sInstrumentConnection(tInstrumentLabel lbl = "", tInstrumentUniqueID id = "", eConnectionInterface itf = eConnectionInterface::kUnknown)
				: Label(lbl), UniqueID(id), Interface(itf)
			{}
			inline bool operator==(const sInstrumentConnection& other) const
			{
				return (UniqueID == other.UniqueID);
			}
			inline bool isEmpty() const
			{
				return (UniqueID.empty() && eConnectionInterface::kUnknown == Interface);
			}
		};
		typedef std::vector<sInstrumentConnection>	tInstrumentList;		// List of instrument names
		enum class eSerialBaudrate {		// Serial baudrate
			k2400		= 2400,
			k4800		= 4800,
			k9600		= 9600,
			k19200		= 19200,
			k28800		= 28800,
			k38400		= 38400,
			k57600		= 57600,
			k115200		= 115200
		};
		enum class eSerialDataBits {		// Number of data bits for the serial communication
			k7	= 7,
			k8	= 8
		};
		enum class eSerialParity {			// Parity configuration for the serial communication
			kNone, kOdd, kEven
		};
		enum class eSerialStopBits {		// Stop bit configuration for the serial communication
			k1,
			k2
		};
		enum class eSerialHandshake {		// Handshake configuration for the serial communication
			kNone,
			kXonXoff, // Software Xon/Xoff Handshake
			kHardware // Hardware CTS RTS Handshake
		};
		struct sMFC3045SerialPortSettings {		// Sets serial port settings
			eSerialBaudrate		Baudrate;			// [bd]	Speed of the serial communication
			eSerialDataBits		DataBits;			// [-]	Number of transmitted bits per packet
			eSerialParity		Parity;				// [-]	Parity configuration of the serial communication
			eSerialStopBits		StopBits;			// [-]	Stop bit configuration of the serial communication
			eSerialHandshake	Handshake;			// [-]	Handshake configuration of the serial communication
			sMFC3045SerialPortSettings()
				: Baudrate(eSerialBaudrate::k9600),
				DataBits(eSerialDataBits::k8),
				Parity(eSerialParity::kNone),
				StopBits(eSerialStopBits::k1),
				Handshake(eSerialHandshake::kNone)
			{}
			bool operator==(const sMFC3045SerialPortSettings & other) const
			{ return (	Baudrate == other.Baudrate &&
						DataBits == other.DataBits && Parity == other.Parity &&
						StopBits == other.StopBits && Handshake == other.Handshake);}
			bool operator!=(const sMFC3045SerialPortSettings & other) const
			{ return !operator==(other); }
		};
		enum class eInstrumentType {		// Type of the instrument
			kMFC3045,
			kPT2026,
			kUnknown
		};
		struct sInstrumentInformation {		// Returns information about the current instrument
			sInstrumentConnection	InstrumentConnection;		// [-]	Connection used for this instrument
			eInstrumentType			Type;						// [-]	Type of the instrument
			std::string				SerialNumber;				// [-]	Serial Number of the instrument
			std::string				Description;				// [-]	Description of the instrument (can contain firmware versions, options, etc...)
			tISO8601DateTime		CalibrationDate;			// [-]	Date of the last calibration in the ISO8601 YYYY-mm-dd format (UTC Date)
			sInstrumentInformation ()
				: InstrumentConnection(), Type(eInstrumentType::kUnknown), SerialNumber(""), Description(""), CalibrationDate("")
			{}
		};
		struct sInstrumentConfiguration {	// Returns information about the current configuration of the instrument
			F64	AcquisitionPeriod;				// [s] - NMR signal acquisition period
			sInstrumentConfiguration()
				: AcquisitionPeriod(DPqNaN)
			{}
		};
		struct sFCAInformation { // FCA peripheral information
			std::string				SerialNumber;		// [-]		Serial Number of the peripheral
			std::string				Description;		// [-]		Description of the peripheral
			U16						FCAFwVersion;		// [-]		Firmware version of the peripheral, if exists. If none => 0
			sFCAInformation()
				: SerialNumber(""), Description(""), FCAFwVersion(0)
			{}
		};

		struct sProbeInformation {			// Returns information about the current selected probe
			U8						ProbeNumber;		// [-]		Number of the probe
			F64						FreqMin;			// [MHz]	Minimal frequency of the probe
			F64						FreqMax;			// [MHz]	Maximum frequency of the probe
			sProbeInformation()
				: ProbeNumber(0), FreqMin(DPqNaN), FreqMax(DPqNaN)
			{}
		};
		enum class eProbeArrayType {		// Type of the Probe-Array
			kMFC3048,
			kMFC9046,
			kMFC9146,
			kUnknown
		};
		struct sProbeArrayInformation {			// Returns information about the current probe-array
			eProbeArrayType						Type;				// [-]		Type of the Probe-Array
			std::string							SerialNumber;		// [-]		Serial Number of the Probe-Array
			std::string							Description;		// [-]		Description of the Probe-Array
			tISO8601DateTime					NormalizationDate;	// [-]		Date of the last normalization in the ISO8601 YYYY-mm-dd format (UTC Date)
			F64									FreqMin;			// [MHz]	Minimal frequency of the probe-array
			F64									FreqMax;			// [MHz]	Maximum frequency of the probe-array
			F64									FreqCentral;		// [MHz]	Central frequency of the probe-array
			F64									GyroRatio;			// [MHz/T]	Gyromagnetic ratio. Should be proton gyro ratio. Results are normalized with this value
			std::vector<sProbeInformation>		ProbeList;			//			List of the contained probes
			sFCAInformation						FCAInfo;			//			FCA peripheral information if exists
			sProbeArrayInformation()
				:	Type(eProbeArrayType::kUnknown),
					SerialNumber(""),
					Description(""),
					NormalizationDate(""),
					FreqMin(DPqNaN),
					FreqMax(DPqNaN),
					FreqCentral(DPqNaN),
					GyroRatio(DPqNaN),
					ProbeList(),
					FCAInfo()
			{}
			size_t NbProbes() const
			{ return ProbeList.size(); }
			void clear()
			{
				Type= eProbeArrayType::kUnknown;
				SerialNumber.clear();
				Description.clear();
				NormalizationDate.clear();
				FreqMin = DPqNaN;
				FreqMax = DPqNaN;
				FreqCentral = DPqNaN;
				GyroRatio = DPqNaN;
				ProbeList.clear();
				FCAInfo = sFCAInformation();
			}
		};
		enum class eOperatingMode {
			kIdle,					// Instrument is not operating
			kSearch,				// Instrument performs a unique search operation
			kMeasure				// Instrument performs measurement according to probe selection
		};
		enum class eRepeatMode {
			kSingle,		//  Operates a single time
			kContinuous		//  Operates continuously until requested to stop
		};
		enum class eMeasureSelectionMode {
			kProbeArray,	// Probe-array is selected. Multiple probes are used
			kSingleProbe	// A single probe in the probe-array is selected
		};
		struct sMeasureProbeSelection
		{
			eMeasureSelectionMode		SelectionMode;	// [-]	Whether we use a single or multiple probes for the measure operation
			U8							ProbeNumber;	// [-]	Number of the selected probe, when SelectionMode is kSingleProbe
			sMeasureProbeSelection()
				: SelectionMode(eMeasureSelectionMode::kProbeArray), ProbeNumber(0)
			{}
			sMeasureProbeSelection(eMeasureSelectionMode SelMode, U8 PrbNb)
				: SelectionMode(SelMode), ProbeNumber(PrbNb)
			{}
			bool operator==(const sMeasureProbeSelection & other) const
			{ return (SelectionMode == other.SelectionMode && ProbeNumber == other.ProbeNumber); }
			bool operator!=(const sMeasureProbeSelection & other) const
			{ return !operator==(other); }
		};
		struct sInstrumentState {				// Controls the state of the instrument, with associated operating parameters
			eOperatingMode			OperatingMode;				// [-]		Requested operating mode
			eRepeatMode				RepeatMode;					// [-]		Requested repeat mode
			sMeasureProbeSelection	ProbeSelection;				// [-]		Probe selection (relevant only for measurement operations)
			F64						FreqMin;					// [MHz]	Minimal frequency used for the specified mode
			F64						FreqMax;					// [MHz]	Maximum frequency used for the specified mode
			U32						NbAveragedMeasurements;		// [-]		Number of repeated measurements used for the average. Limited value from 1 to 1500 for MFC2046 and from 2 to 1500 for MFC3045.
			sInstrumentState()
				: OperatingMode(eOperatingMode::kIdle), RepeatMode(eRepeatMode::kSingle), ProbeSelection(sMeasureProbeSelection()), FreqMin(DPqNaN), FreqMax(DPqNaN), NbAveragedMeasurements(1)
			{}
		};
		enum class eInstrumentState {
			kWaitingForPA,		// There is currently no connected probe-array
			kReady,				// There is a connected probe-array which is currently not operating
			kSearching,			// The instrument is currently searching
			kMeasuring			// The instrument is currently measuring
		};
		struct sInstrumentStatus {		// Reflects the current instrument status
			eInstrumentState	State;				// [-]		Current instrument state
			U8					SearchProgress;		// [%]		Progress of the search (if any) from 0 to 100%
			sInstrumentStatus()
				: State(eInstrumentState::kWaitingForPA), SearchProgress(0)
			{}
			sInstrumentStatus(eInstrumentState St, U8 SPR)
				: State(St), SearchProgress(SPR)
			{}
		};
		enum class eRemoteBoxTrigger {
			kRemoteBoxTrigger		// Informs that Remote box start button has been pressed
		};
		enum class eRemoteBoxBusyLedState {
			kRemoteBusyLedOff,				// Remote Led is turned OFF
			kRemoteBusyLedMeasuring,		// Remote Led is turned ON when measuring and turned OFF otherwise
			kRemoteBusyLedOffEndMeas,		// Remote Led is turned ON immediately and turned OFF automatically with the end of measurement
			kRemoteBusyLedOn,				// Remote Led is turned ON
			kRemoteBusyLedBlinkSlow,		// Remote Led is blinking slowly (about 1Hz)
			kRemoteBusyLedBlinkFast			// Remote Led is blinking quickly (about 2Hz)
		};
		struct sSearchResult {
			F64		Freq;			// [MHz]	Frequency returned by the search operation
			F32		StdDev;			// [ppm]	Standard deviation, relatively to Freq
			sSearchResult()
				: Freq(DPqNaN), StdDev(SPqNaN)
			{}
		};
		struct sMeasurement {
			U8			ProbeNumber;			// [-]		Number of the probe for this measurement
			F64			Freq;					// [MHz]	Measured frequency
			U32			NbValidMeasurements;	// [-]		Number of valid measurements (among the requested number of averaged measurements)
			F32			StdDev;					// [ppm]	Standard deviation, relatively to Freq
			sMeasurement()
				: ProbeNumber(0), Freq(DPqNaN), NbValidMeasurements(0), StdDev(SPqNaN)
			{}
		};
		typedef std::vector<sMeasurement>		tMeasurementList;
		struct sMeasurementResults {
			U64					Timestamp;			// [ms]		Measurement time, relative to startup of the instrument
			tMeasurementList	MeasurementList;	// [-]		List of the obtained measurements
			sMeasurementResults()
				: Timestamp(0), MeasurementList(tMeasurementList())
			{}
		};
		enum class eNormalizationAccess {
			kReadRAM,		// Read the current Normalization table from the temporary memory
			kReadROM,		// Read the Normalization table from the PA EEPROM
			kWriteRAM,		// Write a new Normalization table to the temporary memory
			kWriteROM		// Write a new Normalization table to the PA EEPROM
		};
		typedef F32		tNormalizationCorrection;					// [MHz]	Frequency error correction to be applied for normalization of a probe. Fcorrected = Fmeasured + FCorrection
		typedef std::vector<tNormalizationCorrection>	tNormalizationTable;	// [MHz]	List of normalization correction values for each probe of the probe-array.
		struct sNormalization {
			eNormalizationAccess	Access;				// Whether the normalization is accessed in read or write
			tNormalizationTable		Table;				// Table of correction values according to the reference
			sNormalization()
				: Access(eNormalizationAccess::kReadROM), Table()
			{}
		};
		enum class eErrorLevel {
			kFatal,			// Fatal error. Should not occur, program may crash
			kError,			// Error. The requested operation cannot keep executing
			kWarning		// Warning. An error occured but requested operation will continue to execute
		};
		struct sError {
			eErrorLevel		Level;				// Error level
			I32				Code;				// Code ID of the error. Positive values are warnings, negative values are errors
			std::string		Description;		// Description of the error
			std::string		Source;				// Information about the source of the error
			sError()
				: Level(eErrorLevel::kError), Code(0), Description(""), Source("")
			{}
			sError(eErrorLevel Lvl, I32 Cde, std::string Dsc, std::string Src)
				: Level(Lvl), Code(Cde), Description(Dsc), Source(Src)
			{}
		};
		struct sProbeArrayAngle {
			bool			AngleSensorAvailable;	// [-]		Whether an angle sensor is available or not
			F32				Angle;					// [deg]	Current measured angular position of the probe-array
			sProbeArrayAngle()
				: AngleSensorAvailable(false), Angle(SPqNaN)
			{}
		};
		enum class eLogLevel {
			kCritical,		// An error that causes the software to crash
			kDebug,			// Message for debug purpose
			kInfo,			// General information message that may remain logged in release program
			kWarning		// Something wrong happened but doesn't prevent the software to keep running
		};
		struct sLogMessage {
			eLogLevel		Level;		// Message level
			std::string		Context;	// Message context ("Module.Sub-module.sub-module...")
			std::string		Message;	// Message
			std::string		Location;	// Origin of the message (If any. e.g. "FileName:line")
			sLogMessage()
				: Level(eLogLevel::kDebug)
			{}
			sLogMessage(const eLogLevel & rLevel, const std::string & rContext, const std::string & rMessage, const std::string & rLocation)
				: Level(rLevel), Context(rContext), Message(rMessage), Location(rLocation)
			{}
		};


		//----------------------------------------------------------------------//
		//							Dump Utilities								//
		//----------------------------------------------------------------------//
		static inline std::string l_Dump(const tInstrumentList & rInsList)
		{
			std::ostringstream l_ss;
			for (auto l_i = rInsList.begin(); l_i != rInsList.end(); l_i++)
			{
				l_ss << l_Dump(*l_i);
			}
			return l_ss.str();
		}
		static inline std::string l_Dump(const sInstrumentConnection & rInsConnection)
		{
			std::ostringstream l_ss;
			l_ss << rInsConnection.Label << "(" << rInsConnection.UniqueID << ")";
			l_ss << (	rInsConnection.Interface == eConnectionInterface::kUSB ? " (USB)" :
						rInsConnection.Interface == eConnectionInterface::kEthernet ? " (Eth)" :
						rInsConnection.Interface == eConnectionInterface::kSerial ? " (Serial)" : "(Unkn)") << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sMFC3045SerialPortSettings & rSerialSett)
		{
			std::ostringstream l_ss;
			l_ss << "Baudrate: " << (unsigned long)rSerialSett.Baudrate << " | ";
			l_ss << "DataBits: " << (unsigned long)rSerialSett.DataBits << " | ";
			l_ss << "Parity: " <<	((rSerialSett.Parity == eSerialParity::kNone) ? "None" :
									(rSerialSett.Parity == eSerialParity::kOdd) ? "Odd" :
									(rSerialSett.Parity == eSerialParity::kEven) ? "Even" : "") << " | ";
			l_ss << "StopBits: " <<	((rSerialSett.StopBits == eSerialStopBits::k1) ? "1" :
									(rSerialSett.StopBits == eSerialStopBits::k2) ? "2" : "") << " | ";
			l_ss << "Handshake: " <<	((rSerialSett.Handshake == eSerialHandshake::kNone) ? "None" :
										(rSerialSett.Handshake == eSerialHandshake::kXonXoff) ? "XonXoff" :
										(rSerialSett.Handshake == eSerialHandshake::kHardware) ? "Hardware" : "") << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sInstrumentInformation & rInsInfo)
		{
			std::ostringstream l_ss;
			l_ss << "Instrument Information: Type: " << ((rInsInfo.Type == eInstrumentType::kMFC3045) ? "MFC3045" :
								(rInsInfo.Type == eInstrumentType::kPT2026) ? "PT2026" : " ") << " | ";
			l_ss << "Serial Number: " << rInsInfo.SerialNumber << " | ";
			l_ss << "Calibration Date: " << rInsInfo.CalibrationDate << " | ";
			l_ss << "Connection: " << ((rInsInfo.InstrumentConnection.Interface == eConnectionInterface::kUSB) ? "USB" :
									(rInsInfo.InstrumentConnection.Interface == eConnectionInterface::kSerial) ? "Serial" :
									(rInsInfo.InstrumentConnection.Interface == eConnectionInterface::kEthernet) ? "Ethernet" : "Unknown") << std::endl;
			l_ss << "Description: " << rInsInfo.Description << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sInstrumentConfiguration & rInsConf)
		{
			std::ostringstream l_ss;
			l_ss << "Acquisition Period: " << rInsConf.AcquisitionPeriod << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sProbeArrayInformation & rPAInfo)
		{
			std::ostringstream l_ss;
			l_ss << "Serial Number:      " << rPAInfo.SerialNumber << std::endl;
			l_ss << "Description:        " << rPAInfo.Description << std::endl;
			l_ss << "Normalization Date: " << rPAInfo.NormalizationDate << std::endl;
			l_ss << "FreqMin:            " << rPAInfo.FreqMin << std::endl;
			l_ss << "FreqMax:            " << rPAInfo.FreqMax << std::endl;
			l_ss << "FreqCentral:        " << rPAInfo.FreqCentral << std::endl;
			l_ss << "GyroRatio:          " << rPAInfo.GyroRatio << std::endl;
			l_ss << "ProbeList: (No. probes = " << rPAInfo.NbProbes() << ")" << std::endl;
			for (auto l_it = rPAInfo.ProbeList.begin(); l_it != rPAInfo.ProbeList.end(); l_it++)
			{
				l_ss << " " << (unsigned int)l_it->ProbeNumber << " | " << l_it->FreqMin << " -> " << l_it->FreqMax << std::endl;
			}
			return l_ss.str();
		}
		static inline std::string l_Dump(const sProbeArrayAngle & rPAAngle)
		{
			std::ostringstream l_ss;
			if (rPAAngle.AngleSensorAvailable)
				l_ss << "Angle: " << rPAAngle.Angle << std::endl;
			else
				l_ss << "No PA angle sensor" << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sInstrumentState & rInsState)
		{
			std::ostringstream l_ss;
			l_ss << "OperatingMode:          " << (	rInsState.OperatingMode == eOperatingMode::kIdle ? "Idle" :
													rInsState.OperatingMode == eOperatingMode::kSearch ? "Search" :
													rInsState.OperatingMode == eOperatingMode::kMeasure ? "Measure" : "") << std::endl;
			l_ss << "RepeatMode:             " << (	rInsState.RepeatMode == eRepeatMode::kSingle ? "Single" :
													rInsState.RepeatMode == eRepeatMode::kContinuous ? "Continuous" : "") << std::endl;
			if (rInsState.ProbeSelection.SelectionMode == eMeasureSelectionMode::kProbeArray)
				l_ss << "Probe Array selected" << std::endl;
			else
				l_ss << "Selected probe:         " << (unsigned int)rInsState.ProbeSelection.ProbeNumber << std::endl;
			l_ss << "FreqMin:                " << rInsState.FreqMin << std::endl;
			l_ss << "FreqMax:                " << rInsState.FreqMax << std::endl;
			l_ss << "NbAveragedMeasurements: " << (unsigned int)rInsState.NbAveragedMeasurements << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sInstrumentStatus & rInsStatus)
		{
			std::ostringstream l_ss;
			l_ss << "State: " << (	rInsStatus.State == eInstrumentState::kWaitingForPA ? "WaitingForPA" :
									rInsStatus.State == eInstrumentState::kReady ? "Ready" :
									rInsStatus.State == eInstrumentState::kSearching ? "Searching" :
									rInsStatus.State == eInstrumentState::kMeasuring ? "Measuring" : "") << " | ";
			l_ss << "SearchProgress: " << (unsigned int)rInsStatus.SearchProgress << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sSearchResult & rSearchResult)
		{
			std::ostringstream l_ss;
			l_ss << "Frequency: " << rSearchResult.Freq << " | ";
			l_ss << "StdDev: " << rSearchResult.StdDev << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sMeasurementResults & rMeasResult)
		{
			std::ostringstream l_ss;
			l_ss << "Timestamp: " << rMeasResult.Timestamp << std::endl;
			for (auto l_it = rMeasResult.MeasurementList.begin(); l_it != rMeasResult.MeasurementList.end(); l_it++)
				l_ss << " ProbeNo.: " << (unsigned int)l_it->ProbeNumber << " Freq: " << l_it->Freq << " Nb.Valid: " << (unsigned int)l_it->NbValidMeasurements << " SDev: " << l_it->StdDev << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const tISO8601DateTime & rDateTime)
		{
			std::ostringstream l_ss;
			l_ss << "Date: " << rDateTime << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sNormalization & rNormalization)
		{
			std::ostringstream l_ss;
			l_ss << (	rNormalization.Access == eNormalizationAccess::kReadRAM ? "ReadRAM Access" :
						rNormalization.Access == eNormalizationAccess::kReadROM ? "ReadROM Access" :
						rNormalization.Access == eNormalizationAccess::kWriteRAM ? "WriteRAM Access" :
						rNormalization.Access == eNormalizationAccess::kWriteROM ? "WriteROM Access" : "") << std::endl;
			for (auto l_it = rNormalization.Table.begin(); l_it != rNormalization.Table.end(); l_it++)
				l_ss << *l_it << std::endl;
			l_ss << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const eRemoteBoxBusyLedState & rRBBusyLedState)
		{
			std::ostringstream l_ss;
			l_ss << (	rRBBusyLedState == eRemoteBoxBusyLedState::kRemoteBusyLedOff ? "Remote busy led off" :
						rRBBusyLedState == eRemoteBoxBusyLedState::kRemoteBusyLedMeasuring ? "Remote busy led on when measuring" :
						rRBBusyLedState == eRemoteBoxBusyLedState::kRemoteBusyLedOffEndMeas ? "Remote busy led clears after measuring" :
						rRBBusyLedState == eRemoteBoxBusyLedState::kRemoteBusyLedOn ? "Remote busy led on" :
						rRBBusyLedState == eRemoteBoxBusyLedState::kRemoteBusyLedBlinkSlow ? "Remote busy led blink slow" :
						rRBBusyLedState == eRemoteBoxBusyLedState::kRemoteBusyLedBlinkFast ? "Remote busy led blink fast" : "") << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sError & rError)
		{
			std::ostringstream l_ss;
			l_ss << "Error : " << (	rError.Level == eErrorLevel::kFatal ? "Fatal " :
									rError.Level == eErrorLevel::kError ? "Error " :
									rError.Level == eErrorLevel::kWarning ? "Warning " : "");
			l_ss << rError.Code << " : " << rError.Description << " | @ " << rError.Source << std::endl;
			return l_ss.str();
		}
		static inline std::string l_Dump(const sLogMessage & rLogMessage)
		{
			std::ostringstream l_ss;
			l_ss << "Level: " << (	rLogMessage.Level == eLogLevel::kCritical ? "Critical" :
									rLogMessage.Level == eLogLevel::kDebug ? "Debug" :
									rLogMessage.Level == eLogLevel::kInfo ? "Info" :
									rLogMessage.Level == eLogLevel::kWarning ? "Warning" : "") << std::endl;
			l_ss << "Context: " << rLogMessage.Context << std::endl;
			l_ss << "Message: \"" << rLogMessage.Message << "\"" << std::endl;
			l_ss << "@ " << rLogMessage.Location << std::endl;
			return l_ss.str();
		}



		friend class CGUICatalogInterface;
		friend class CInstrManCatalogInterface;

	protected:
		//----------------------------------------------------------------------//
		//								Variables								//
		//----------------------------------------------------------------------//
		//------------------------------------------//
		// Connection
		CVariable<tInstrumentList>				vInstrumentList;
		CVariable<sInstrumentConnection>		vInstrumentCurrent;
		CVariable<sMFC3045SerialPortSettings>	vMFC3045SerialPortSettings;
		//------------------------------------------//
		// Instrument / Probe-Array information
		CVariable<sInstrumentInformation>		vInstrumentInformation;
		CVariable<sInstrumentConfiguration>		vInstrumentConfiguration;
		CVariable<sProbeArrayInformation>		vProbeArrayInformation;
		CVariable<sProbeArrayAngle>				vProbeArrayAngle;
		//------------------------------------------//
		// Instrument / Probe-Array control and status
		CVariable<sInstrumentState>				vInstrumentState;
		CVariable<sInstrumentStatus>			vInstrumentStatus;
		CVariable<eRemoteBoxTrigger>			vRemoteBoxTriggerEvent;
		CVariable<eRemoteBoxBusyLedState>		vRemoteBoxBusyLedState;
		//------------------------------------------//
		// Results
		CVariable<sSearchResult>				vSearchResult;
		CVariable<sMeasurementResults>			vMeasurementResults;
		//------------------------------------------//
		// Others
		CVariable<sNormalization>				vNormalizationRequest;
		CVariable<sNormalization>				vNormalizationReturned;
		CVariable<sError>						vError;
		CVariable<sLogMessage>					vLogMessage;

	public:
		CInstrumentCatalog()
			: vRemoteBoxTriggerEvent(eRemoteBoxTrigger::kRemoteBoxTrigger), vRemoteBoxBusyLedState(eRemoteBoxBusyLedState::kRemoteBusyLedOff)
		{}

	}; // class CInstrumentCatalog

	class CGUICatalogInterface
	{
	public:
		//----------------------------------------------------------------------//
		//							Controls and Observers						//
		//----------------------------------------------------------------------//
		//------------------------------------------//
		// Connection
		CObserver<CInstrumentCatalog::tInstrumentList>				oInstrumentList;
		CControl<CInstrumentCatalog::sInstrumentConnection>			cInstrumentCurrent;
		CControl<CInstrumentCatalog::sMFC3045SerialPortSettings>	cMFC3045SerialPortSettings;
		//------------------------------------------//
		// Instrument / Probe-Array information
		CObserver<CInstrumentCatalog::sInstrumentInformation>		oInstrumentInformation;
		CObserver<CInstrumentCatalog::sInstrumentConfiguration>		oInstrumentConfiguration;
		CObserver<CInstrumentCatalog::sProbeArrayInformation>		oProbeArrayInformation;
		CObserver<CInstrumentCatalog::sProbeArrayAngle>				oProbeArrayAngle;
		//------------------------------------------//
		// Instrument / Probe-Array control and status
		CControl<CInstrumentCatalog::sInstrumentState>				cInstrumentState;
		CObserver<CInstrumentCatalog::sInstrumentStatus>			oInstrumentStatus;
		CObserver<CInstrumentCatalog::eRemoteBoxTrigger>			oRemoteBoxTriggerEvent;
		CControl<CInstrumentCatalog::eRemoteBoxBusyLedState>		cRemoteBoxBusyLedState;
		//------------------------------------------//
		// Results
		CObserver<CInstrumentCatalog::sSearchResult>				oSearchResult;
		CObserver<CInstrumentCatalog::sMeasurementResults>			oMeasurementResults;
		//------------------------------------------//
		// Others
		CControl<CInstrumentCatalog::sNormalization>				cNormalizationRequest;
		CObserver<CInstrumentCatalog::sNormalization>				oNormalizationReturned;
		CObserver<CInstrumentCatalog::sError>						oError;
		CObserver<CInstrumentCatalog::sLogMessage>					oLogMessage;

		CGUICatalogInterface(CInstrumentCatalog & rCatalog)
			: oInstrumentList(rCatalog.vInstrumentList),
			cInstrumentCurrent(rCatalog.vInstrumentCurrent),
			cMFC3045SerialPortSettings(rCatalog.vMFC3045SerialPortSettings),
			oInstrumentInformation(rCatalog.vInstrumentInformation),
			oInstrumentConfiguration(rCatalog.vInstrumentConfiguration),
			oProbeArrayInformation(rCatalog.vProbeArrayInformation),
			oProbeArrayAngle(rCatalog.vProbeArrayAngle),
			cInstrumentState(rCatalog.vInstrumentState),
			oInstrumentStatus(rCatalog.vInstrumentStatus),
			oRemoteBoxTriggerEvent(rCatalog.vRemoteBoxTriggerEvent),
			cRemoteBoxBusyLedState(rCatalog.vRemoteBoxBusyLedState),
			oSearchResult(rCatalog.vSearchResult),
			oMeasurementResults(rCatalog.vMeasurementResults),
			cNormalizationRequest(rCatalog.vNormalizationRequest),
			oNormalizationReturned(rCatalog.vNormalizationReturned),
			oError(rCatalog.vError),
			oLogMessage(rCatalog.vLogMessage)
		{}
	};

	class CInstrManCatalogInterface
	{
	public:
		//----------------------------------------------------------------------//
		//							Controls and Observers						//
		//----------------------------------------------------------------------//
		//------------------------------------------//
		// Connection
		CControl<CInstrumentCatalog::tInstrumentList>				cInstrumentList;
		CObserver<CInstrumentCatalog::sInstrumentConnection>		oInstrumentCurrent;
		CObserver<CInstrumentCatalog::sMFC3045SerialPortSettings>	oMFC3045SerialPortSettings;
		//------------------------------------------//
		// Instrument / Probe-Array information
		CControl<CInstrumentCatalog::sInstrumentInformation>		cInstrumentInformation;
		CControl<CInstrumentCatalog::sInstrumentConfiguration>		cInstrumentConfiguration;
		CControl<CInstrumentCatalog::sProbeArrayInformation>		cProbeArrayInformation;
		CControl<CInstrumentCatalog::sProbeArrayAngle>				cProbeArrayAngle;
		//------------------------------------------//
		// Instrument / Probe-Array control and status
		CObserver<CInstrumentCatalog::sInstrumentState>				oInstrumentState;
		CControl<CInstrumentCatalog::sInstrumentStatus>				cInstrumentStatus;
		CControl<CInstrumentCatalog::eRemoteBoxTrigger>				cRemoteBoxTriggerEvent;
		CObserver<CInstrumentCatalog::eRemoteBoxBusyLedState>		oRemoteBoxBusyLedState;
		//------------------------------------------//
		// Results
		CControl<CInstrumentCatalog::sSearchResult>					cSearchResult;
		CControl<CInstrumentCatalog::sMeasurementResults>			cMeasurementResults;
		//------------------------------------------//
		// Others
		CObserver<CInstrumentCatalog::sNormalization>				oNormalizationRequest;
		CControl<CInstrumentCatalog::sNormalization>				cNormalizationReturned;
		CControl<CInstrumentCatalog::sError>						cError;
		CControl<CInstrumentCatalog::sLogMessage>					cLogMessage;

		CInstrManCatalogInterface(CInstrumentCatalog & rCatalog)
			: cInstrumentList(rCatalog.vInstrumentList),
			oInstrumentCurrent(rCatalog.vInstrumentCurrent),
			oMFC3045SerialPortSettings(rCatalog.vMFC3045SerialPortSettings),
			cInstrumentInformation(rCatalog.vInstrumentInformation),
			cInstrumentConfiguration(rCatalog.vInstrumentConfiguration),
			cProbeArrayInformation(rCatalog.vProbeArrayInformation),
			cProbeArrayAngle(rCatalog.vProbeArrayAngle),
			oInstrumentState(rCatalog.vInstrumentState),
			cInstrumentStatus(rCatalog.vInstrumentStatus),
			cRemoteBoxTriggerEvent(rCatalog.vRemoteBoxTriggerEvent),
			oRemoteBoxBusyLedState(rCatalog.vRemoteBoxBusyLedState),
			cSearchResult(rCatalog.vSearchResult),
			cMeasurementResults(rCatalog.vMeasurementResults),
			oNormalizationRequest(rCatalog.vNormalizationRequest),
			cNormalizationReturned(rCatalog.vNormalizationReturned),
			cError(rCatalog.vError),
			cLogMessage(rCatalog.vLogMessage)
		{}
	};
} // namespace MTL
