///	\file			MFC2x46Device.h
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#pragma once

// Standard includes
#include <string>
// Personal includes
#include "VISAInstrument.h"
#include "PT2026.h"
#include "MFCDevice.h"
#include "InstrumentCatalog.h"
#include "MFCController.h"
#include "Helpers.h"


using namespace MTL;
using namespace MTL::Instrument;

namespace MTL {

	//--------------------------------------------------------------------------//
	//					MFC2x46 Device (PT2026 + MFC9046 or MFC9146)
	//--------------------------------------------------------------------------//

	class CMFC2x46Device: public CMFCDevice
	{
	//--------------------------------------------------------------------------//
	//								Attributes
	//--------------------------------------------------------------------------//
	private:
		CMFCController &								m_rMFCController;
		CPT2026Instrument								m_PT2026;
		CExtendedChannelList							m_ChanList;
		SCPI::tChannel									m_PAChannel;
		U8												m_NbOfSearchHeads = 0;
		U8												m_NbOfMeasHeads = 0;
		U8												m_NbOfMeasurementsToFetch = 0;
		CInstrumentCatalog::sInstrumentState			m_LastConfigured;

		std::string										m_DefaultPAConfScriptFilePath;		// Path of the default PA configuration script (if no specific file exists)
		std::string										m_CustomPAConfScriptDirectory;		// Directory containing PA specific configuration scripts

	//constants:
		const int			SEARCH_POLLING_PERIOD_MS		= 100;
		const int			MEAS_POLLING_PERIOD_MS			= 50;
		const eUnits		UNITS							= kMHz;

	//--------------------------------------------------------------------------//
	//							Constructor / Destructor
	//--------------------------------------------------------------------------//
	public:
		CMFC2x46Device (CMFCController &	rMFCController, CVISAResourceManager & rRM, tResourceName Rsrc,
					   const std::string & rDefaultPAConfScriptFilePath,
					   const std::string & rCustomPAConfScriptDir);
		virtual ~CMFC2x46Device ();

	//--------------------------------------------------------------------------//
	//							Utilities
	//--------------------------------------------------------------------------//
	private:
		bool l_UtcToStrDateISO8601(const std::time_t & rUtcDate, std::string & rStrDate);
		bool l_ExecuteConfigurationScript (U16 Model, U32 SN, CMFC2x46Device::sMFCDeviceError & rError);
		bool l_ConfigurationChanged (CInstrumentCatalog::sInstrumentState lastcfg, CInstrumentCatalog::sInstrumentState newcfg);
		bool l_ConfigureForOperation (const CInstrumentCatalog::sInstrumentState & rInstrState, CMFC2x46Device::sMFCDeviceError & rError);
		void l_ComputeAveraging (const std::vector<tFlux> & rFluxList, F64 & rReturnedAveraging, F64 & rReturnedStdDev);
	public:
		bool AbortInstrument ();
		bool IsInstrumentConnected();

	//--------------------------------------------------------------------------//
	//							Available Operations
	//--------------------------------------------------------------------------//
	public:
		sMFCDeviceError Connect (CInstrumentCatalog::sInstrumentInformation & rInstrInfo);
		sMFCDeviceError ScanPA (	CInstrumentCatalog::sProbeArrayInformation & rPAInfo,
									CInstrumentCatalog::sInstrumentConfiguration & rInstrConfig, bool & rPaInfoChanged);
		sMFCDeviceError Search (const CInstrumentCatalog::sInstrumentState & rInstrState);
		sMFCDeviceError Measure (const CInstrumentCatalog::sInstrumentState & rInstrState);
		sMFCDeviceError NewNormalizationRequest (	const CInstrumentCatalog::sNormalization & rNewNormalizationRequest,
													CInstrumentCatalog::sNormalization & rReturnedNormalization);
		sMFCDeviceError GetNormalizationDate(CInstrumentCatalog::tISO8601DateTime & rNormDate);
		sMFCDeviceError SetRBBusyLedState (const CInstrumentCatalog::eRemoteBoxBusyLedState & rRBBusyLedState);
		sMFCDeviceError RemoteBoxButtonStatus (bool & rRemoteStartPressed);
		sMFCDeviceError ChangeSerialParmsOnConnectedInstr(const CInstrumentCatalog::sMFC3045SerialPortSettings & rNewSerialSetting); //We don't use it for M2x46Device

	}; // class
} // namespace MTL
