///	\file			MFC3045Device.h
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#pragma once

// Standard includes
#include <string>
// Personal includes
#include "VISAInstrument.h"
#include "MFC3045.h"
#include "MFCDevice.h"
#include "InstrumentCatalog.h"
#include "MFCController.h"
#include "Helpers.h"

using namespace MTL;
using namespace MTL::Instrument;

// Status Register selection
#define SetST1		(0b1)
#define SetST2		(0b10)
#define SetST3		(0b100)
#define SetST4		(0b1000)
#define SetST5		(0b10000)
#define SetST6		(0b100000)
#define SetSTALL	(0xFF)

union uStatusRegSel {
	U8	SelSTReg = 0;
	struct sStatusRegSel {
		// Contains a 1-byte status register 1. The MFC3045 uses the following bits:
		U8		ST1 : 1;	// bit[0]: Sel Status Register 1
		U8		ST2 : 1;	// bit[1]: Sel Status Register 2
		U8		ST3 : 1;	// bit[2]: Sel Status Register 3
		U8		ST4 : 1;	// bit[3]: Sel Status Register 4
		U8		ST5 : 1;	// bit[4]: Sel Status Register 5
		U8		ST6 : 1;	// bit[5]: Sel Status Register 6
		U8		ALLST : 2;	// bit[6-7] Sel All Status Register
	} StRegSelBit;
};

namespace MTL {

	//--------------------------------------------------------------------------//
	//								MFC3045 Device
	//--------------------------------------------------------------------------//

	class CMFC3045Device: public CMFCDevice
	{
		//--------------------------------------------------------------------------//
		//								Attributes
		//--------------------------------------------------------------------------//
	private:
		CMFCController &							m_rMFCController;
		CMFC3045Instrument							m_MFC3045;
		sSerialPortSettings							m_SerialParms;
		std::string									m_PaSerialNum;
		U8											m_NbProbesPA;

		uStatusByte1								m_RegStatus1;
		uStatusByte2								m_RegStatus2;
		uStatusByte3								m_RegStatus3;
		uStatusByte4								m_RegStatus4;
		uStatusByte5								m_RegStatus5;
		uStatusByte6								m_RegStatus6;
		std::string									m_LastCmdError;

		//constants:
			const int			SEARCH_POLLING_PERIOD_MS		= 100;
			const int			MEAS_POLLING_PERIOD_MS			= 100;

		//--------------------------------------------------------------------------//
		//							Constructor / Destructor
		//--------------------------------------------------------------------------//
	public:
		CMFC3045Device (	CMFCController &	rMFCController, CVISAResourceManager & rRM,
							tResourceName Rsrc, CInstrumentCatalog::sMFC3045SerialPortSettings & rSerialSetting);
		virtual ~CMFC3045Device ();

		//--------------------------------------------------------------------------//
		//								Utilities
		//--------------------------------------------------------------------------//
	private:
		bool l_UtcToStrDateISO8601(const std::time_t & rUtcDate, std::string & rStrDate);
		bool l_ParseSerialSettings (	const CInstrumentCatalog::sMFC3045SerialPortSettings & rSerialSetting,
										sSerialPortSettings & rVisaSerialSetting);
		bool l_ReadSomeStatusReg(const uStatusRegSel & rRdSelStatus);
		bool l_SerialCommunicationSuccess();
		bool l_CheckPAConnected(bool & rPAPresent);
		bool l_ConfigurationChanged (CInstrumentCatalog::sInstrumentState lastcfg, CInstrumentCatalog::sInstrumentState newcfg);
		bool l_ConfigureForOperation (const CInstrumentCatalog::sInstrumentState & rInstrState, sMFCDeviceError & rError);
		bool l_ComputeMaxMeasTimeout(U64 & rMaxTimeOutMs);
		void l_ComputeAveraging (const std::vector<U32> & rFluxListdHz, F64 & rReturnedAveragingMHz, F64 & rReturnedStdDevppm);
		bool l_WriteNormalizationRAM(const CInstrumentCatalog::tNormalizationTable & rNewNormalizationTable, sMFCDeviceError & rError);

	public:
		bool AbortInstrument ();
		bool IsInstrumentConnected();


		//--------------------------------------------------------------------------//
		//							Available Operations
		//--------------------------------------------------------------------------//
	public:	
		sMFCDeviceError Connect (CInstrumentCatalog::sInstrumentInformation & rInstrInfo);
		sMFCDeviceError ScanPA (	CInstrumentCatalog::sProbeArrayInformation & rPAInfo,
									CInstrumentCatalog::sInstrumentConfiguration & rInstrConfig, bool &rPaInfoChanged);
		sMFCDeviceError Search (const CInstrumentCatalog::sInstrumentState & rInstrState);
		sMFCDeviceError Measure (const CInstrumentCatalog::sInstrumentState & rInstrState);
		sMFCDeviceError NewNormalizationRequest (	const CInstrumentCatalog::sNormalization & rNewNormalizationRequest,
													CInstrumentCatalog::sNormalization & rReturnedNormalization);
		sMFCDeviceError GetNormalizationDate(CInstrumentCatalog::tISO8601DateTime & rNormDate);
		sMFCDeviceError SetRBBusyLedState (const CInstrumentCatalog::eRemoteBoxBusyLedState & rRBBusyLedState);
		sMFCDeviceError RemoteBoxButtonStatus (bool & rRemoteStartPressed);
		sMFCDeviceError ChangeSerialParmsOnConnectedInstr(const CInstrumentCatalog::sMFC3045SerialPortSettings & rNewSerialSetting);

	}; // class
} // namespace MTL
