///	\file			ThreadedMailbox.h
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#pragma once

#include <mutex>
#include <thread>
#include <condition_variable>
#include <queue>
#include <chrono>
#include <memory>
#include <atomic>

namespace MTL {
	namespace ThreadedMailbox {

		class CSemaphore
		{
		private:
			std::mutex m_mutex;
			std::condition_variable m_condition;
			unsigned long m_count = 0; // Initialized as locked.

		public:
			/// \brief Reset semaphore object.
			void reset() {
				std::unique_lock<decltype(m_mutex)> lock(m_mutex);
				m_count = 0;
			}

			/// \brief Raise a semaphore.
			void notify() {
				std::unique_lock<decltype(m_mutex)> lock(m_mutex);
				++m_count;
				m_condition.notify_one();
			}

			/// \brief Wait for a semaphore.
			void wait() {
				std::unique_lock<decltype(m_mutex)> lock(m_mutex);
				while (!m_count) // Handle spurious wake-ups.
					m_condition.wait(lock);
				--m_count;
			}

			/// \brief Wait for a semaphore for a limited time.
			bool wait_for(std::chrono::milliseconds timeout_ms) {
				std::cv_status l_Stat = std::cv_status::no_timeout;
				std::unique_lock<decltype(m_mutex)> lock(m_mutex);
				while (!m_count && l_Stat == std::cv_status::no_timeout) // Handle spurious wake-ups.
					l_Stat = m_condition.wait_for(lock, timeout_ms);
				if (l_Stat == std::cv_status::no_timeout)
					--m_count;
				return (l_Stat == std::cv_status::no_timeout);
			}

			/// \brief Check whether a semaphore has been raised.
			bool try_wait() {
				std::unique_lock<decltype(m_mutex)> lock(m_mutex);
				if (m_count) {
					--m_count;
					return true;
				}
				return false;
			}
		};

		template <typename MsgT>
		class CMailBox
		{
		private:
			std::queue<MsgT>	m_MsgQ;
			std::mutex			m_Lock;
			CSemaphore			m_Sem;

		public:
			bool empty()
			{
				m_Lock.lock();
				bool l_em = m_MsgQ.empty();
				m_Lock.unlock();
				return l_em;
			}
			void clear()
			{
				m_Lock.lock();
				std::queue<MsgT> l_Empty;
				std::swap(m_MsgQ, l_Empty);
				m_Sem.reset();
				m_Lock.unlock();
			}
			void push(MsgT pMsg)
			{
				m_Lock.lock();
				m_MsgQ.push(pMsg);
				m_Sem.notify();
				m_Lock.unlock();
			}
			void pop(MsgT & rV)
			{
				m_Sem.wait();
				m_Lock.lock();
				if (m_MsgQ.empty())
					return;
				rV = m_MsgQ.front();
				m_MsgQ.pop();
				m_Lock.unlock();
			}
			bool try_pop(MsgT & rV)
			{
				bool l_Success = false;
				m_Lock.lock();
				if (m_Sem.try_wait())
				{
					l_Success = true;
					rV = m_MsgQ.front();
					m_MsgQ.pop();
				}
				m_Lock.unlock();
				return l_Success;
			}
		};

		class CTimer
		{
		private:
			std::thread					m_Thread;		// Thread object
			std::mutex					m_Lock;			// Lock onto public interface
			std::chrono::milliseconds	m_Period_ms;
			bool						m_SingleShot;
			CSemaphore					m_AbortSem;
			std::function<void(void)>	m_NotifyEvent;

			//-----------------------------------------//
			// Thread Context
			//-----------------------------------------//
		public:
			// Constructor as: CTimer timer(Function)
			CTimer(const std::function<void(void)> & rTimerEventNotifierFunction)
				: m_NotifyEvent(rTimerEventNotifierFunction)
			{}
			// Constructor as: CTimer timer(&Class::method, &ClassObj)
			template<class Fn, class... Args>
			CTimer(Fn&& rTimerEventNotifierFunction, Args&&... Ax)
				: m_NotifyEvent(std::bind(rTimerEventNotifierFunction, Ax...))
			{}
			virtual ~CTimer()
			{
				Stop();
			}
		private:
			void ThreadMain()
			{
				while (1)
				{
					// Wait for timer period to elapse
					bool l_Aborted = m_AbortSem.wait_for(m_Period_ms);

					// If requested to abort, quit the thread without notification
					if (l_Aborted)
						return;

					// Notify event
					m_NotifyEvent();

					// Check if we continue
					if (m_SingleShot)
						return;					// If single shot, stop thread
				}
			}
			//-----------------------------------------//
			// Caller Context
			//-----------------------------------------//
		public:
			bool StartSingleShot(std::chrono::milliseconds Delay_ms)
			{
				std::lock_guard<decltype(m_Lock)> l_lg(m_Lock);

				if (m_Thread.joinable())		// If thread is already running
					return false;

				m_SingleShot = true;
				m_Period_ms = Delay_ms;
				m_AbortSem.reset();
				m_Thread = std::thread(&CTimer::ThreadMain, this);
				return true;
			}
			bool Start(std::chrono::milliseconds Period_ms)
			{
				std::lock_guard<decltype(m_Lock)> l_lg(m_Lock);

				if (m_Thread.joinable())		// If thread is already running
					return false;

				m_SingleShot = false;
				m_Period_ms = Period_ms;
				m_AbortSem.reset();
				m_Thread = std::thread(&CTimer::ThreadMain, this);
				return true;
			}
			void Stop()
			{
				std::lock_guard<decltype(m_Lock)> l_lg(m_Lock);

				if (m_Thread.joinable())		// If worker thread is running
				{
					m_AbortSem.notify();			// Notify thread to stop
					m_Thread.join();				// Wait for worker thread (and its timer thread) to be stopped
				}
			}
		};


		//-----------------------------------------//
		//-----------------------------------------//
		// Class
		//-----------------------------------------//
		//-----------------------------------------//
		template <typename UserMsgT>
		class CThreadedMailBox
		{
			//-----------------------------------------//
			// Types
			//-----------------------------------------//
		protected:
			enum eEventType { kExit, kUserMsg };
		private:
			struct sMailBoxMsg {
				eEventType		Evt;
				UserMsgT		UserMsg;
				sMailBoxMsg()
				{}
				sMailBoxMsg(eEventType EvtType, const UserMsgT & rMsg)
					: Evt(EvtType), UserMsg(rMsg)
				{}
				sMailBoxMsg(eEventType EvtType)
					: Evt(EvtType)
				{}
			};
		protected:
			//-----------------------------------------//
			// Attributes
			//-----------------------------------------//
		private:
			CMailBox<sMailBoxMsg>	m_MailBox;
			std::thread				m_Thread;
			std::atomic<bool>		m_RequestedToStop;

			//-----------------------------------------//
			// Caller context
			//-----------------------------------------//
		protected:
			void l_SendMsg(const UserMsgT & rMsg);
		public:
			CThreadedMailBox()
				: m_RequestedToStop(false)
			{}
			void ThreadStart();
			void ThreadStop();
		protected:
			bool l_IsRequestedToStop();
			//-----------------------------------------//
			// Worker Thread context
			//-----------------------------------------//
		private:
			void l_ThreadMain();
		protected:
			virtual void l_ThreadTask(eEventType EvtType, UserMsgT * pMsg) = 0;
		};


		//-----------------------------------------//
		//-----------------------------------------//
		// Implementation
		//-----------------------------------------//
		//-----------------------------------------//
		template <typename UserMsgT>
		void CThreadedMailBox<UserMsgT>::ThreadStart()
		{
			if (!m_Thread.joinable())		// If thread is not already running
			{
				m_MailBox.clear();
				m_RequestedToStop = false;
				m_Thread = std::thread(&CThreadedMailBox::l_ThreadMain, this);
			}
		}

		template <typename UserMsgT>
		void CThreadedMailBox<UserMsgT>::ThreadStop()
		{
			m_RequestedToStop = true;
			m_MailBox.push(sMailBoxMsg(eEventType::kExit));									// Request worker thread to stop
			if (m_Thread.joinable() && std::this_thread::get_id() != m_Thread.get_id())		// If worker thread is running
				m_Thread.join();															// Wait for worker thread to be stopped
		}

		template <typename UserMsgT>
		void CThreadedMailBox<UserMsgT>::l_SendMsg(const UserMsgT & rMsg)
		{
			m_MailBox.push(sMailBoxMsg(eEventType::kUserMsg, rMsg));	// Push our copy to our message queue
		}

		template <typename UserMsgT>
		bool CThreadedMailBox<UserMsgT>::l_IsRequestedToStop()
		{
			return m_RequestedToStop;
		}

		template <typename UserMsgT>
		void CThreadedMailBox<UserMsgT>::l_ThreadMain()
		{
			sMailBoxMsg Msg;
			bool l_RunWorkerThread = true;
			while (l_RunWorkerThread)
			{
				m_MailBox.pop(Msg);
				switch (Msg.Evt)
				{
				case eEventType::kExit:
					l_RunWorkerThread = false;		// Request this thread to stop
					l_ThreadTask(Msg.Evt, nullptr);	// Call thread task for the last time
					break;
				case eEventType::kUserMsg:
					l_ThreadTask(Msg.Evt, &Msg.UserMsg);		// Forward user message
					break;
				}
			}
		}

	} // namespace ThreadedMailbox
} // namespace MTL
