///	\file			CPT2026PeripheralROM.cpp
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

// Standard includes

// Personal includes
#include "CPT2026PeripheralROM.h"
#include "Helpers.h"

// External tools includes
#include "date.h"

using namespace MTL;

// Utilities
#define UNUSED_VAR(x)	(void)x;	// Prevents from unereferenced variable when compiling in release

//----------------------------------------------------------------------//
//							Conversion Utilities						//
//----------------------------------------------------------------------//
// From Human representation	(Atomic size: 1 byte, Endianness: Little endian)
// To local machine representation
// Both IEEE 754 format
inline U16 l_PeripheralBufferToU16(const U8* it)
{
#ifdef _WIN64
	return static_cast<U16>(*it) << 8 | static_cast<U16>(*(it + 1));
#endif
}
inline I16 l_PeripheralBufferToI16(const U8* it)
{
#ifdef _WIN64
	return static_cast<I16>(static_cast<U16>(*it) << 8 | static_cast<U16>(*(it + 1)));
#endif
}
inline U32 l_PeripheralBufferToU32(const U8* it)
{
#ifdef _WIN64
	return static_cast<U32>(*it) << 24 | static_cast<U32>(*(it + 1)) << 16 | static_cast<U32>(*(it + 2)) << 8 | static_cast<U32>(*(it + 3));
#endif
}
inline I32 l_PeripheralBufferToI32(const U8* it)
{
#ifdef _WIN64
	return static_cast<I32>(static_cast<U32>(*it) << 24 | static_cast<U32>(*(it + 1)) << 16 | static_cast<U32>(*(it + 2)) << 8 | static_cast<U32>(*(it + 3)));
#endif
}
inline F32 l_PeripheralBufferToF32(const U8* it)
{
#ifdef _WIN64
	U32 l_swap =	static_cast<U32>(it[0]) << 24 |
					static_cast<U32>(it[1]) << 16 |
					static_cast<U32>(it[2]) << 8 |
					static_cast<U32>(it[3]);
	return *reinterpret_cast<F32*>(&l_swap);
#endif
}
inline F64 l_PeripheralBufferToF64(const U8* it)
{
#ifdef _WIN64
	U64 l_swap =	static_cast<U64>(it[0]) << 56 |
					static_cast<U64>(it[1]) << 48 |
					static_cast<U64>(it[2]) << 40 |
					static_cast<U64>(it[3]) << 32 |
					static_cast<U64>(it[4]) << 24 |
					static_cast<U64>(it[5]) << 16 |
					static_cast<U64>(it[6]) << 8 |
					static_cast<U64>(it[7]);
	return *reinterpret_cast<F64*>(&l_swap);
#endif
}
inline void l_U16ToPeripheralBuffer(U16 n, U8* it)
{
	*it++ = static_cast<U8>(n >> 8);
	*it++ = static_cast<U8>(n);
}
inline void l_U16ArrayToPeripheralBuffer(const std::vector<U16> & rna, U8* it)
{
	for (auto & l_it : rna)
	{
		l_U16ToPeripheralBuffer(l_it, it);
		it += sizeof(U16);
	}
}
inline void l_F64ToPeripheralBuffer(F64 f, U8* it)
{
#ifdef _WIN64
	U64 l_swap =	static_cast<U64>(reinterpret_cast<U8*>(&f)[0]) << 56 |
					static_cast<U64>(reinterpret_cast<U8*>(&f)[1]) << 48 |
					static_cast<U64>(reinterpret_cast<U8*>(&f)[2]) << 40 |
					static_cast<U64>(reinterpret_cast<U8*>(&f)[3]) << 32 |
					static_cast<U64>(reinterpret_cast<U8*>(&f)[4]) << 24 |
					static_cast<U64>(reinterpret_cast<U8*>(&f)[5]) << 16 |
					static_cast<U64>(reinterpret_cast<U8*>(&f)[6]) << 8 |
					static_cast<U64>(reinterpret_cast<U8*>(&f)[7]);
	std::memcpy(&*it, &l_swap, sizeof(F64));
#endif
}
inline void l_F64ArrayToPeripheralBuffer(std::vector<F64> fa, U8* it)
{
	for (auto & l_it : fa)
		l_F64ToPeripheralBuffer(l_it, it += sizeof(F64));
}
inline void l_F32ToPeripheralBuffer(F32 f, U8* it)
{
#ifdef _WIN64
	U32 l_swap =	static_cast<U32>(reinterpret_cast<U8*>(&f)[0]) << 24 |
					static_cast<U32>(reinterpret_cast<U8*>(&f)[1]) << 16 |
					static_cast<U32>(reinterpret_cast<U8*>(&f)[2]) << 8 |
					static_cast<U32>(reinterpret_cast<U8*>(&f)[3]);
	std::memcpy(&*it, &l_swap, sizeof(F32));
#endif
}
inline void l_F32ArrayToPeripheralBuffer(std::vector<F32> fa, U8* it)
{
	for (auto & l_it : fa)
	{
		l_F32ToPeripheralBuffer(l_it, it);
		it += sizeof(F32);
	}
}

U8 l_CRC(std::vector<U8>::const_iterator beg, std::vector<U8>::const_iterator end)
{
	U8 l_CRC_Val = 0;
	for (; beg != end; beg++)
		l_CRC_Val ^= *beg;
	return l_CRC_Val;
}

//----------------------------------------------------------------------//
//							High Level Methods							//
//----------------------------------------------------------------------//
bool CPT2026PeripheralROM::Parser::Parse()
{
	try
	{
		U16 l_ROMVersion;
		if (!l_HeaderVersion(l_ROMVersion))
			throw false;

		switch (l_ROMVersion)
		{
		case sStandaloneStaticHeader::VERSION:
		{
			m_HeaderType = kStandalone;
			l_ParseStaticHeader(m_StdandaloneHeader);
			l_ParseParmInfo(m_StdandaloneHeader.HEADER_LEN, m_ParmInfoMap);
			break;
		}
		case sEmbeddedStaticHeader::VERSION:
		{
			m_HeaderType = kEmbedded;
			l_ParseStaticHeader(m_EmbeddedHeader);
			l_ParseParmInfo(m_EmbeddedHeader.HEADER_LEN, m_ParmInfoMap);
			break;
		}
		default:
			MTL_Assert(false);
			break;
		}
	}
	catch (bool & rE)
	{
		UNUSED_VAR(rE);
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::HasIdentification(sIdentification & rID)
{
	switch (m_HeaderType)
	{
	case kStandalone:
		rID = m_StdandaloneHeader.Identification;
		return true;
		break;
	default:
		return false;
		break;
	}
}

bool CPT2026PeripheralROM::Parser::ParmAddress(U8 ParmNo, U16 & rParmAddress)
{
	CParmInfoMap::const_iterator l_it = m_ParmInfoMap.find(ParmNo);
	if (m_ParmInfoMap.end() != l_it)
	{
		rParmAddress = l_it->second;
		return true;
	}
	else
	{
		rParmAddress = 0;
		return false;
	}
}

CPT2026PeripheralROM::eHeaderType CPT2026PeripheralROM::Parser::HeaderType()
{
	return m_HeaderType;
}

CPT2026PeripheralROM::sStandaloneStaticHeader CPT2026PeripheralROM::Parser::Header(const CPT2026PeripheralROM::sStandaloneStaticHeader & rHeader)
{
	UNUSED_VAR(rHeader);
	return m_StdandaloneHeader;
}

CPT2026PeripheralROM::sEmbeddedStaticHeader CPT2026PeripheralROM::Parser::Header(const CPT2026PeripheralROM::sEmbeddedStaticHeader & rHeader)
{
	UNUSED_VAR(rHeader);
	return m_EmbeddedHeader;
}

//----------------------------------------------------------------------//
//						Common Type Parsing Methods						//
//----------------------------------------------------------------------//
bool CPT2026PeripheralROM::Parser::l_ParseIdentification(sIdentification & rId, std::vector<U8>::const_iterator Id_beg, std::vector<U8>::const_iterator Id_end)
{
	try
	{
		if (static_cast<size_t>(Id_end - Id_beg) < sIdentification::IDENTIFICATION_LEN)
			throw false;
		// Parse
		rId.ProtocolVersion			= static_cast<U16>(static_cast<U16>(*Id_beg) << 8 | *(Id_beg + 1));
		rId.DeviceType				= static_cast<eDeviceType>(*(Id_beg + 2));
		rId.Model					= static_cast<U16>(static_cast<U16>(*(Id_beg + 3)) << 8 | static_cast<U16>(*(Id_beg + 4)));
		rId.SN						= static_cast<U32>(*(Id_beg + 5)) << 16 | static_cast<U32>(*(Id_beg + 6)) << 8 | *(Id_beg + 7);
		rId.Capabilities			= static_cast<U32>(*(Id_beg + 8)) << 24 | static_cast<U32>(*(Id_beg + 9)) << 16 | static_cast<U32>(*(Id_beg + 10)) << 8 | *(Id_beg + 11);
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::l_HeaderVersion(U16 & rVersion)
{
	try
	{
		rVersion = static_cast<U16>(static_cast<U16>(*m_beg) << 8 | *(m_beg + 1));
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::l_ParseStaticHeader(sStandaloneStaticHeader & rHeader)
{
	try
	{
		// Check version
		if (!l_HeaderVersion(rHeader.ROMVersion) || rHeader.ROMVersion != sStandaloneStaticHeader::VERSION)
			return false;
		// Check CRC
		if (0 != l_CRC(m_beg, m_beg + sStandaloneStaticHeader::HEADER_LEN))
			return false;

		// Parse
		rHeader.HeaderCRC			= *(m_beg + sStandaloneStaticHeader::HEADER_LEN - 1);
		rHeader.ROMCapacity			= static_cast<U16>(static_cast<U16>(*(m_beg + 2)) << 8 | *(m_beg + 3));
		rHeader.ROMSize				= static_cast<U16>(static_cast<U16>(*(m_beg + 4)) << 8 | *(m_beg + 5));
		std::copy(m_beg + 6, m_beg + 6 + sStandaloneStaticHeader::DATE_LEN, rHeader.ROMProgrammingDate);
		if (!l_ParseIdentification(rHeader.Identification, m_beg + 16, m_end))
			return false;
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::l_ParseStaticHeader(sEmbeddedStaticHeader & rHeader)
{
	try
	{
		// Check version
		if (!l_HeaderVersion(rHeader.ROMVersion) || rHeader.ROMVersion != sEmbeddedStaticHeader::VERSION)
			return false;
		// Check CRC
		rHeader.HeaderCRC			= *(m_beg + sEmbeddedStaticHeader::HEADER_LEN - 1);
		if (rHeader.HeaderCRC != l_CRC(m_beg, m_beg + sEmbeddedStaticHeader::HEADER_LEN))
			return false;

		// Parse
		rHeader.ROMCapacity			= static_cast<U16>(static_cast<U16>(*(m_beg + 2)) << 8 | *(m_beg + 3));
		rHeader.ROMSize				= static_cast<U16>(static_cast<U16>(*(m_beg + 4)) << 8 | *(m_beg + 5));
		std::copy(m_beg + 6, m_beg + 6 + sEmbeddedStaticHeader::DATE_LEN, rHeader.ROMProgrammingDate);
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::l_ParseParmInfo(size_t HeaderLen, CParmInfoMap & rParmInfoMap)
{
	try
	{
		std::vector<U8>::const_iterator l_beg = m_beg + HeaderLen;

		// Parse
		U16 l_NoParms = static_cast<U16>(static_cast<U16>(*l_beg) << 8 | *(l_beg + 1));

		l_beg += 2;
		for (U16 l_NoParsedParmInfo = 0; l_NoParsedParmInfo < l_NoParms; l_NoParsedParmInfo++, l_beg += 3)
			rParmInfoMap.emplace(*l_beg, static_cast<U16>(*(l_beg + 1)) << 8 | *(l_beg + 2));	// ParmNo / Address
	}
	catch (...)
	{
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//					Peripheral Type Parsing Methods						//
//----------------------------------------------------------------------//
bool CPT2026PeripheralROM::Parser::ParseManufactureInformation(size_t ParmAddress, sManufactureInformation & rManuInfo)
{
	try
	{
		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		U16 l_ManDateLen = static_cast<U16>(static_cast<U16>(*l_it) << 8 | *(l_it + 1));
		l_it += 2;
		rManuInfo.ManufacturingDate = std::string(l_it, l_it + l_ManDateLen);
		l_it += l_ManDateLen;

		U16 l_DesLen = static_cast<U16>(static_cast<U16>(*l_it) << 8 | *(l_it + 1));
		l_it += 2;
		rManuInfo.Designation = std::string(l_it, l_it + l_DesLen);
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::ParseMultiHeadInfo(size_t ParmAddress, sMultiHeadInfo & rMultiHeadInfo)
{
	try
	{
		rMultiHeadInfo.MultiHeadInfo.clear();

		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		rMultiHeadInfo.StructVersion	= l_PeripheralBufferToU16(&(*l_it));		l_it += sizeof(rMultiHeadInfo.StructVersion);
		if (rMultiHeadInfo.StructVersion != sMultiHeadInfo::STRUCT_VERSION)	return false;

		rMultiHeadInfo.OffsetDAC		= l_PeripheralBufferToU16(&(*l_it));		l_it += sizeof(rMultiHeadInfo.OffsetDAC);

		U8 l_NoHeads = l_it[0];		l_it += sizeof(l_NoHeads);
		for (U8 HeadNo = 0; HeadNo < l_NoHeads; HeadNo++)
		{
			sHeadInfo l_HeadInfo;
			l_HeadInfo.MUXcode = l_it[0];
			l_HeadInfo.Info = l_it[1];
			l_HeadInfo.PhysInfoSetIndex = l_it[2];
			l_it += sHeadInfo::SIZE;
			rMultiHeadInfo.MultiHeadInfo.push_back(l_HeadInfo);
		}
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::ParseHeadList(size_t ParmAddress, tHeadList & rHeadList)
{
	try
	{
		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		rHeadList.clear();
		U8 l_NoHeads = l_it[0];
		l_it++;
		rHeadList = tHeadList(l_it, l_it + l_NoHeads);
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::ParsePwCamPhysInfo(size_t ParmAddress, sPwCamPhysicalInfo & rPhysInfo)
{
	try
	{
		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		rPhysInfo.Version		= l_PeripheralBufferToU16(&(*l_it));			l_it += sizeof(rPhysInfo.Version);
		if (rPhysInfo.Version != sPwCamPhysicalInfo::HANDLED_PW_CAM_PHYSICAL_INFORMATION_VERSION)	return false;
		rPhysInfo.LowFreq		= l_PeripheralBufferToF64(&(*l_it));			l_it += sizeof(rPhysInfo.LowFreq);
		rPhysInfo.NominalFreq	= l_PeripheralBufferToF64(&(*l_it));			l_it += sizeof(rPhysInfo.NominalFreq);
		rPhysInfo.HighFreq		= l_PeripheralBufferToF64(&(*l_it));			l_it += sizeof(rPhysInfo.HighFreq);
		rPhysInfo.GyroRatio		= l_PeripheralBufferToF64(&(*l_it));			l_it += sizeof(rPhysInfo.GyroRatio);
		U8 l_NormalizationDateSize = l_it[0];
		l_it++;
		rPhysInfo.NormalizationDate = std::string(l_it, l_it + l_NormalizationDateSize);
		l_it += l_NormalizationDateSize;
		U8 l_NbCorrections = l_it[0];		l_it += sizeof(l_NbCorrections);
		for (U8 HeadNo = 0; HeadNo < l_NbCorrections; HeadNo++, l_it += sizeof(F32))
			rPhysInfo.NormalizationTable.push_back(l_PeripheralBufferToF32(&(*l_it)));
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::ParsePwPhysicalInfo(size_t ParmAddress, sPwPhysicalInformation & rPhysInfo, std::vector<U8>::const_iterator & rpBuffer)
{
	try
	{
		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		U16 l_StructVersion = l_PeripheralBufferToU16(&(*l_it));			l_it += sizeof(U16);
		if (l_StructVersion != sPwPhysicalInformation::HANDLED_PWP_PHYSICAL_INFORMATION_VERSION)	return false;
		rPhysInfo.MinFreq			= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.MaxFreq			= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.MinIF				= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.MaxIF				= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.FreqMultiplier	= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.GyroRatio			= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.PulsePeriod		= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.T2				= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.SampleLength		= l_PeripheralBufferToF64(&(*l_it));	l_it += sizeof(F64);
		rPhysInfo.DeadTime			= l_it[0];								l_it += 1;
		// Get Noise Level curve
		U16 l_NoNoiseLevelElements = l_PeripheralBufferToU16(&(*l_it));				l_it += sizeof(U16);
		if (l_NoNoiseLevelElements > sPwPhysicalInformation::MAX_CURVE_SIZE)		return false;
		for (U16 l_PointNo = 0; l_PointNo < l_NoNoiseLevelElements; l_PointNo++, l_it += sizeof(U16))
			rPhysInfo.NoiseLevelCurve.push_back(l_PeripheralBufferToU16(&(*l_it)));
		// Get Tuning curve
		U16 l_NoTuningElements = l_PeripheralBufferToU16(&(*l_it));					l_it += sizeof(U16);
		if (l_NoTuningElements > sPwPhysicalInformation::MAX_CURVE_SIZE)			return false;
		for (U16 l_PointNo = 0; l_PointNo < l_NoTuningElements; l_PointNo++, l_it += sizeof(U16))
			rPhysInfo.TuningCurve.push_back(l_PeripheralBufferToU16(&(*l_it)));
		// Get Matching curve
		U16 l_NoMatchingElements = l_PeripheralBufferToU16(&(*l_it));				l_it += sizeof(U16);
		if (l_NoMatchingElements > sPwPhysicalInformation::MAX_CURVE_SIZE)			return false;
		for (U16 l_PointNo = 0; l_PointNo < l_NoMatchingElements; l_PointNo++, l_it += sizeof(U16))
			rPhysInfo.MatchingCurve.push_back(l_PeripheralBufferToU16(&(*l_it)));
		// Get pulse width curve
		U16 l_NoPulseWidthElements = l_PeripheralBufferToU16(&(*l_it));				l_it += sizeof(U16);
		if (l_NoPulseWidthElements > sPwPhysicalInformation::MAX_CURVE_SIZE)		return false;
		for (U16 l_PointNo = 0; l_PointNo < l_NoPulseWidthElements; l_PointNo++, l_it += sizeof(U16))
			rPhysInfo.PulseWidthCurve.push_back(l_PeripheralBufferToU16(&(*l_it)));
		// Get Freq step curve
		U16 l_NoFreqStepElements = l_PeripheralBufferToU16(&(*l_it));				l_it += sizeof(U16);
		if (l_NoFreqStepElements > sPwPhysicalInformation::MAX_CURVE_SIZE)			return false;
		for (U16 l_PointNo = 0; l_PointNo < l_NoFreqStepElements; l_PointNo++, l_it += sizeof(U16))
			rPhysInfo.FreqStepCurve.push_back(l_PeripheralBufferToU16(&(*l_it)));

		rpBuffer = l_it;
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::ParsePwPhysInfoSet(size_t ParmAddress, sPwPhysicalInformationSet & rPwPhysInfoSet)
{
	try
	{
		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		U8 l_NoSets = l_it[0];
		l_it++;
		for (U8 SetNo = 0; SetNo < l_NoSets; SetNo++)
		{
			sPwPhysicalInformation* pPwPhysInfo = new sPwPhysicalInformation;
			if (ParsePwPhysicalInfo(static_cast<size_t>(l_it - m_beg), *pPwPhysInfo, l_it))
				rPwPhysInfoSet.Attach(pPwPhysInfo);
			else
				return false;
		}
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::ParsePwFcaPhysInfo(size_t ParmAddress, sPwFCAPhysInfo & rPhysInfo)
{
	try
	{
		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		U16 l_StructVersion = l_PeripheralBufferToU16(&(*l_it));			l_it += 2;
		if (l_StructVersion != sPwFCAPhysInfo::HANDLED_PW_FCA_PHYSICAL_INFORMATION_VERSION)	return false;
		rPhysInfo.MinFreq			= l_PeripheralBufferToF64(&(*l_it));	l_it += 8;
		rPhysInfo.MaxFreq			= l_PeripheralBufferToF64(&(*l_it));	l_it += 8;
		rPhysInfo.MinIF				= l_PeripheralBufferToF64(&(*l_it));	l_it += 8;
		rPhysInfo.MaxIF				= l_PeripheralBufferToF64(&(*l_it));	l_it += 8;
		rPhysInfo.FreqMultiplier	= l_PeripheralBufferToF64(&(*l_it));	l_it += 8;
		rPhysInfo.VoltageNoise		= l_PeripheralBufferToU16(&(*l_it));	l_it += 2;
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool CPT2026PeripheralROM::Parser::ParseHallInfo(size_t ParmAddress, sHallInformation & rHallInfo)
{
	try
	{
		std::vector<U8>::const_iterator l_it = m_beg + ParmAddress;

		U16 l_StructVersion = l_PeripheralBufferToU16(&(*l_it));			l_it += 2;
		if (l_StructVersion != sHallInformation::STRUCT_VERSION)	return false;
		// Get the state
		rHallInfo.State = *l_it;											l_it += 1;
		// Get the range
		rHallInfo.Range_mT = l_PeripheralBufferToU16(&(*l_it));				l_it += 2;
		// Get the tolerance
		rHallInfo.Tolerance = *l_it;										l_it += 1;
		// Get the calibration coefficients
		rHallInfo.Calibration.B0X	= l_PeripheralBufferToU16(&(*l_it));	l_it += 2;
		rHallInfo.Calibration.B0Y	= l_PeripheralBufferToU16(&(*l_it));	l_it += 2;
		rHallInfo.Calibration.B0Z	= l_PeripheralBufferToU16(&(*l_it));	l_it += 2;
		rHallInfo.Calibration.G0X	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.G0Y	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.G0Z	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.T0	= l_PeripheralBufferToU16(&(*l_it));	l_it += 2;
		rHallInfo.Calibration.CT0X	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.CT0Y	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.CT0Z	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.CTGX	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.CTGY	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
		rHallInfo.Calibration.CTGZ	= l_PeripheralBufferToF32(&(*l_it));	l_it += 4;
	}
	catch (...)
	{
		return false;
	}
	return true;
}

std::vector<U8> CPT2026PeripheralROM::Builder::l_Raw(const CPT2026PeripheralROM::sIdentification & rId)
{
	std::vector<U8> l_Raw(rId.IDENTIFICATION_LEN, 0);		// Allocate fixed size memory
	auto l_it = l_Raw.begin();

	*l_it++ = static_cast<U8>(rId.ProtocolVersion >> 8);						// Fill data
	*l_it++ = static_cast<U8>(rId.ProtocolVersion);
	*l_it++ = static_cast<U8>(rId.DeviceType);
	*l_it++ = static_cast<U8>(rId.Model >> 8);
	*l_it++ = static_cast<U8>(rId.Model);
	*l_it++ = static_cast<U8>(rId.SN >> 16);
	*l_it++ = static_cast<U8>(rId.SN >> 8);
	*l_it++ = static_cast<U8>(rId.SN);
	*l_it++ = static_cast<U8>(rId.Capabilities >> 24);
	*l_it++ = static_cast<U8>(rId.Capabilities >> 16);
	*l_it++ = static_cast<U8>(rId.Capabilities >> 8);
	*l_it++ = static_cast<U8>(rId.Capabilities);

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::sHallInformation & rHallInfo)
{
	std::vector<U8> l_Raw;

	l_U16ToPeripheralBuffer(rHallInfo.STRUCT_VERSION, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_Raw.push_back(rHallInfo.State);
	l_U16ToPeripheralBuffer(rHallInfo.Range_mT, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_Raw.push_back(rHallInfo.Tolerance);
	l_U16ToPeripheralBuffer(rHallInfo.Calibration.B0X, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_U16ToPeripheralBuffer(rHallInfo.Calibration.B0Y, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_U16ToPeripheralBuffer(rHallInfo.Calibration.B0Z, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.G0X, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.G0Y, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.G0Z, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_U16ToPeripheralBuffer(rHallInfo.Calibration.T0, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.CT0X, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.CT0Y, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.CT0Z, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.CTGX, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.CTGY, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));
	l_F32ToPeripheralBuffer(rHallInfo.Calibration.CTGZ, &*l_Raw.insert(l_Raw.end(), sizeof(F32), 0));

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::sManufactureInformation & rManInfo)
{
	std::vector<U8> l_Raw;

	l_U16ToPeripheralBuffer(static_cast<U16>(rManInfo.ManufacturingDate.length()), &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_Raw.insert(l_Raw.end(), rManInfo.ManufacturingDate.begin(), rManInfo.ManufacturingDate.end());
	l_U16ToPeripheralBuffer(static_cast<U16>(rManInfo.Designation.length()), &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_Raw.insert(l_Raw.end(), rManInfo.Designation.begin(), rManInfo.Designation.end());

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::sMultiHeadInfo & rMultHeadInfo)
{
	std::vector<U8> l_Raw;

	l_Raw.push_back(rMultHeadInfo.STRUCT_VERSION >> 8);
	l_Raw.push_back(rMultHeadInfo.STRUCT_VERSION);
	l_Raw.push_back(static_cast<U8>(rMultHeadInfo.OffsetDAC >> 8));
	l_Raw.push_back(static_cast<U8>(rMultHeadInfo.OffsetDAC));
	l_Raw.push_back(static_cast<U8>(rMultHeadInfo.MultiHeadInfo.size()));
	for (auto & l_it : rMultHeadInfo.MultiHeadInfo)
	{
		l_Raw.push_back(l_it.MUXcode);
		l_Raw.push_back(l_it.Info);
		l_Raw.push_back(l_it.PhysInfoSetIndex);
	}

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::tHeadList & rHeadList)
{
	std::vector<U8> l_Raw;

	l_Raw.push_back(static_cast<U8>(rHeadList.size()));
	l_Raw.insert(l_Raw.end(), rHeadList.begin(), rHeadList.end());

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::sPwCamPhysicalInfo & rPwCamPhysInfo)
{
	std::vector<U8> l_Raw;

	l_U16ToPeripheralBuffer(rPwCamPhysInfo.HANDLED_PW_CAM_PHYSICAL_INFORMATION_VERSION, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_F64ToPeripheralBuffer(rPwCamPhysInfo.LowFreq, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwCamPhysInfo.NominalFreq, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwCamPhysInfo.HighFreq, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwCamPhysInfo.GyroRatio, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_Raw.push_back(static_cast<U8>(rPwCamPhysInfo.NormalizationDate.length()));
	l_Raw.insert(l_Raw.end(), rPwCamPhysInfo.NormalizationDate.begin(), rPwCamPhysInfo.NormalizationDate.end());
	U8 l_NbNormElements = static_cast<U8>(rPwCamPhysInfo.NormalizationTable.size());
	l_Raw.push_back(static_cast<U8>(l_NbNormElements));
	l_F32ArrayToPeripheralBuffer(rPwCamPhysInfo.NormalizationTable, &*l_Raw.insert(l_Raw.end(), l_NbNormElements * sizeof(F32), 0));

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::sPwFCAPhysInfo & rPwFCAPhysInfo)
{
	std::vector<U8> l_Raw;

	l_U16ToPeripheralBuffer(rPwFCAPhysInfo.HANDLED_PW_FCA_PHYSICAL_INFORMATION_VERSION, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_F64ToPeripheralBuffer(rPwFCAPhysInfo.MinFreq, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwFCAPhysInfo.MaxFreq, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwFCAPhysInfo.MinIF, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwFCAPhysInfo.MaxIF, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwFCAPhysInfo.FreqMultiplier, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_U16ToPeripheralBuffer(rPwFCAPhysInfo.VoltageNoise, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::sPwPhysicalInformation & rPwPhysInfo)
{
	std::vector<U8> l_Raw;

	l_U16ToPeripheralBuffer(rPwPhysInfo.HANDLED_PWP_PHYSICAL_INFORMATION_VERSION, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.MinFreq, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.MaxFreq, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.MinIF, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.MaxIF, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.FreqMultiplier, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.GyroRatio, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.PulsePeriod, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.T2, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_F64ToPeripheralBuffer(rPwPhysInfo.SampleLength, &*l_Raw.insert(l_Raw.end(), sizeof(F64), 0));
	l_Raw.push_back(rPwPhysInfo.DeadTime);
	U16 l_NbElements = static_cast<U16>(rPwPhysInfo.NoiseLevelCurve.size());
	l_U16ToPeripheralBuffer(l_NbElements, &*l_Raw.insert(l_Raw.end(), sizeof(l_NbElements), 0));
	if (0 != l_NbElements)
		l_U16ArrayToPeripheralBuffer(rPwPhysInfo.NoiseLevelCurve, &*l_Raw.insert(l_Raw.end(), l_NbElements * sizeof(U16), 0));
	l_NbElements = static_cast<U16>(rPwPhysInfo.TuningCurve.size());
	l_U16ToPeripheralBuffer(l_NbElements, &*l_Raw.insert(l_Raw.end(), sizeof(l_NbElements), 0));
	if (0 != l_NbElements)
		l_U16ArrayToPeripheralBuffer(rPwPhysInfo.TuningCurve, &*l_Raw.insert(l_Raw.end(), l_NbElements * sizeof(U16), 0));
	l_NbElements = static_cast<U16>(rPwPhysInfo.MatchingCurve.size());
	l_U16ToPeripheralBuffer(l_NbElements, &*l_Raw.insert(l_Raw.end(), sizeof(l_NbElements), 0));
	if (0 != l_NbElements)
		l_U16ArrayToPeripheralBuffer(rPwPhysInfo.MatchingCurve, &*l_Raw.insert(l_Raw.end(), l_NbElements * sizeof(U16), 0));
	l_NbElements = static_cast<U16>(rPwPhysInfo.PulseWidthCurve.size());
	l_U16ToPeripheralBuffer(l_NbElements, &*l_Raw.insert(l_Raw.end(), sizeof(l_NbElements), 0));
	if (0 != l_NbElements)
		l_U16ArrayToPeripheralBuffer(rPwPhysInfo.PulseWidthCurve, &*l_Raw.insert(l_Raw.end(), l_NbElements * sizeof(U16), 0));
	l_NbElements = static_cast<U16>(rPwPhysInfo.FreqStepCurve.size());
	l_U16ToPeripheralBuffer(l_NbElements, &*l_Raw.insert(l_Raw.end(), sizeof(l_NbElements), 0));
	if (0 != l_NbElements)
		l_U16ArrayToPeripheralBuffer(rPwPhysInfo.FreqStepCurve, &*l_Raw.insert(l_Raw.end(), l_NbElements * sizeof(U16), 0));

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::Raw(const CPT2026PeripheralROM::sPwPhysicalInformationSet & rPwPhysInfoSet)
{
	std::vector<U8> l_Raw;

	l_Raw.push_back(static_cast<U8>(rPwPhysInfoSet.size()));		// Number of PwParmInfo structs in the set
	for (CPT2026PeripheralROM::sPwPhysicalInformationSet::size_type l_index = 0; l_index < rPwPhysInfoSet.size(); l_index++)
	{
		std::vector<U8> l_RawPhPhysInfo = Raw(rPwPhysInfoSet[l_index]);
		l_Raw.insert(l_Raw.end(), l_RawPhPhysInfo.begin(), l_RawPhPhysInfo.end());
	}

	return l_Raw;
}

CPT2026PeripheralROM::CParmInfoMap CPT2026PeripheralROM::Builder::l_BuildParmInfo(const tParmContentInfo & rParmContent, U16 BaseAddress, size_t & rParmDataSize)
{
	rParmDataSize = 0;
	CParmInfoMap l_ParmInfo;
	for (auto & l_it : rParmContent)
	{
		l_ParmInfo.insert(std::pair<U8, U16>(l_it.ParmNumber, BaseAddress));
		BaseAddress += static_cast<U16>(l_it.rParmData.size());
		rParmDataSize += l_it.rParmData.size();
	}
	return l_ParmInfo;
}

CPT2026PeripheralROM::CParmInfoMap CPT2026PeripheralROM::Builder::l_BuildParmInfo(const tParmContentInfo & rParmContent, U16 BaseAddress, std::vector<U8> & rGeneratedParmData)
{
	rGeneratedParmData.clear();
	CParmInfoMap l_ParmInfo;
	for (auto & l_it : rParmContent)
	{
		l_ParmInfo.insert(std::pair<U8, U16>(l_it.ParmNumber, BaseAddress));
		BaseAddress += static_cast<U16>(l_it.rParmData.size());
		rGeneratedParmData.insert(rGeneratedParmData.end(), l_it.rParmData.begin(), l_it.rParmData.end());
	}
	return l_ParmInfo;
}

std::vector<U8> CPT2026PeripheralROM::Builder::l_Raw(const sStandaloneStaticHeader & rHeader)
{
	std::vector<U8> l_Raw;

	l_Raw.push_back(static_cast<U8>(rHeader.ROMVersion >> 8));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMVersion));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMCapacity >> 8));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMCapacity));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMSize >> 8));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMSize));
	std::memcpy(&*l_Raw.insert(l_Raw.end(), sStandaloneStaticHeader::DATE_LEN, 0), rHeader.ROMProgrammingDate, sStandaloneStaticHeader::DATE_LEN);
	std::vector<U8> l_RawId = CPT2026PeripheralROM::Builder::l_Raw(rHeader.Identification);
	l_Raw.insert(l_Raw.end(), l_RawId.begin(), l_RawId.end());
	l_Raw.push_back(l_CRC(l_Raw.begin(), l_Raw.begin() + l_Raw.size()));

	MTL_Assert(l_Raw.size() == rHeader.HEADER_LEN);

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::l_Raw(const sEmbeddedStaticHeader & rHeader)
{
	std::vector<U8> l_Raw;

	l_Raw.push_back(static_cast<U8>(rHeader.ROMVersion >> 8));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMVersion));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMCapacity >> 8));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMCapacity));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMSize >> 8));
	l_Raw.push_back(static_cast<U8>(rHeader.ROMSize));
	std::memcpy(&*l_Raw.insert(l_Raw.end(), sEmbeddedStaticHeader::DATE_LEN, 0), rHeader.ROMProgrammingDate, sEmbeddedStaticHeader::DATE_LEN);
	l_Raw.push_back(l_CRC(l_Raw.begin(), l_Raw.begin() + l_Raw.size()));

	MTL_Assert(l_Raw.size() == rHeader.HEADER_LEN);

	return l_Raw;
}

std::vector<U8> CPT2026PeripheralROM::Builder::l_Raw(const CParmInfoMap & rParmInfo)
{
	std::vector<U8> l_Raw;

	l_U16ToPeripheralBuffer(static_cast<U16>(rParmInfo.size()), &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));		// NbParms
	for (auto & l_it : rParmInfo)
	{
		l_Raw.push_back(l_it.first);																// Device Parm Number
		l_U16ToPeripheralBuffer(l_it.second, &*l_Raw.insert(l_Raw.end(), sizeof(U16), 0));		// Parm address
	}

	return l_Raw;
}

template <typename HeaderType>
std::vector<U8> CPT2026PeripheralROM::Builder::l_BuildROM(HeaderType Header, const tParmContentInfo & rParmContent)
{
	// Compute sizes
	size_t l_HeaderLen = Header.HEADER_LEN;
	size_t l_ParmInfoLen, l_ParmDataLen;
	l_ParmInfoLen = (l_Raw(l_BuildParmInfo(rParmContent, 0, l_ParmDataLen))).size();		// Build a dummy raw parm info from a dummy base address to get the sizes
	
	// Prepare header
	Header.ROMVersion = Header.VERSION;
	//Header.ROMCapacity must have been previously filled
    Header.ROMSize = static_cast<U16>(l_HeaderLen + l_ParmInfoLen + l_ParmDataLen);
	Header.HeaderCRC = 0;
	std::vector<U8> l_RawHeaderBuff = l_Raw(Header);

	// Prepare parm info
	std::vector<U8> l_ParmDataBuff;
	CParmInfoMap l_ParmInfoMap = l_BuildParmInfo(rParmContent, static_cast<U16>(l_HeaderLen + l_ParmInfoLen), l_ParmDataBuff);
	std::vector<U8> l_RawParmInfoBuff = l_Raw(l_ParmInfoMap);

	// Fill ROM
	std::vector<U8> l_ROM;
	l_ROM.insert(l_ROM.end(), l_RawHeaderBuff.begin(), l_RawHeaderBuff.end());
	l_ROM.insert(l_ROM.end(), l_RawParmInfoBuff.begin(), l_RawParmInfoBuff.end());
	l_ROM.insert(l_ROM.end(), l_ParmDataBuff.begin(), l_ParmDataBuff.end());
	
	return l_ROM;
}

std::vector<U8> CPT2026PeripheralROM::Builder::BuildROM(sStandaloneStaticHeader Header, const tParmContentInfo & rParmContent)
{
	return CPT2026PeripheralROM::Builder::l_BuildROM(Header, rParmContent);
}

std::vector<U8> CPT2026PeripheralROM::Builder::BuildROM(sEmbeddedStaticHeader Header, const tParmContentInfo & rParmContent)
{
	return CPT2026PeripheralROM::Builder::l_BuildROM(Header, rParmContent);
}

void CPT2026PeripheralROM::Builder::Build1226ROM(	std::vector<U8> & rOutputROM,
													const CPT2026PeripheralROM::sManufactureInformation & rManInfo,
													const CPT2026PeripheralROM::sPwPhysicalInformation & rPwPhysInfo,
													const CPT2026PeripheralROM::sHallInformation & rHallInfo)
{
	rOutputROM.clear();

	std::vector<U8> l_RawManInfo = CPT2026PeripheralROM::Builder::Raw(rManInfo);
	std::vector<U8> l_RawPwPhysInfo = CPT2026PeripheralROM::Builder::Raw(rPwPhysInfo);
	std::vector<U8> l_RawHallInfo = CPT2026PeripheralROM::Builder::Raw(rHallInfo);

	CPT2026PeripheralROM::Builder::tParmContentInfo l_ParmInfo =
	{
		{ static_cast<U8>(CPT2026PeripheralROM::e1226ROMParm::kManufactureInformation),		l_RawManInfo },
		{ static_cast<U8>(CPT2026PeripheralROM::e1226ROMParm::kPwPhysicalInformation),		l_RawPwPhysInfo },
		{ static_cast<U8>(CPT2026PeripheralROM::e1226ROMParm::kHallInformation),			l_RawHallInfo }
	};

	CPT2026PeripheralROM::sEmbeddedStaticHeader l_Header;
	//l_Header.ROMVersion			// Auto handled in BuildROM
	l_Header.ROMCapacity = 1024;
	//l_Header.ROMSize				// Auto handled in BuildROM
	std::memcpy(l_Header.ROMProgrammingDate, date::format("%F", std::chrono::system_clock::now()).c_str(), l_Header.DATE_LEN);

	rOutputROM = BuildROM(l_Header, l_ParmInfo);
}

void CPT2026PeripheralROM::Builder::Build9046ROM(std::vector<U8> & rOutputROM,
	const CPT2026PeripheralROM::sIdentification & rIdentification,
	const CPT2026PeripheralROM::sManufactureInformation & rManInfo,
	const CPT2026PeripheralROM::sMultiHeadInfo & rMultiHeadInfo,
	const CPT2026PeripheralROM::tHeadList & rSearchHeadList,
	const CPT2026PeripheralROM::tHeadList & rMeasHeadList,
	const CPT2026PeripheralROM::sPwCamPhysicalInfo & rPwCamPhysInfo,
	const CPT2026PeripheralROM::sPwPhysicalInformationSet & rPwPhysInfoSet)
{
	rOutputROM.clear();

	std::vector<U8> l_RawManInfo = CPT2026PeripheralROM::Builder::Raw(rManInfo);
	std::vector<U8> l_RawMultiHeadInfo = CPT2026PeripheralROM::Builder::Raw(rMultiHeadInfo);
	std::vector<U8> l_RawSearchHeadList = CPT2026PeripheralROM::Builder::Raw(rSearchHeadList);
	std::vector<U8> l_RawMeasHeadList = CPT2026PeripheralROM::Builder::Raw(rMeasHeadList);
	std::vector<U8> l_RawPwCamPhysInfo = CPT2026PeripheralROM::Builder::Raw(rPwCamPhysInfo);
	std::vector<U8> l_RawPwPhysInfoSet = CPT2026PeripheralROM::Builder::Raw(rPwPhysInfoSet);

	CPT2026PeripheralROM::Builder::tParmContentInfo l_ParmInfo =
	{
		{ static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kManufactureInformation),		l_RawManInfo },
		{ static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kMultiHeadInformation),		l_RawMultiHeadInfo },
		{ static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kSearchHeads),				l_RawSearchHeadList },
		{ static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kMeasHeads),					l_RawMeasHeadList },
		{ static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwCamPhysicalInformation),	l_RawPwCamPhysInfo },
		{ static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwPhysicalInformationSet),	l_RawPwPhysInfoSet }
	};

	CPT2026PeripheralROM::sStandaloneStaticHeader l_Header;
	//l_Header.ROMVersion			// Auto handled in BuildROM
	l_Header.ROMCapacity = 8192;
	//l_Header.ROMSize				// Auto handled in BuildROM
	std::memcpy(l_Header.ROMProgrammingDate, date::format("%F", std::chrono::system_clock::now()).c_str(), l_Header.DATE_LEN);
	l_Header.Identification = rIdentification;

	rOutputROM = BuildROM(l_Header, l_ParmInfo);
}
