///	\file			MFCController.cpp
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

// Standard includes
#include <regex>
#include <fstream>		// ifstream to read files
#include <sstream>		// string manipulation

// Personal includes
#include "MFCController.h"
#include "OSDefines.h"

#define LOG(__msg__, __level__,__location__)		l_SendLogMessage(__msg__, __level__, "", __location__ )

#if defined _WIN32
#define DIR_SEP		'\\'
#else
#define DIR_SEP		'/'
#endif


//--------------------------------------------------------------------------//
//								MFC CONTROLLER
//--------------------------------------------------------------------------//
void CMFCController::l_SendLogMessage(const std::string & rMessage, const CInstrumentCatalog::eLogLevel & rLevel, const std::string & rContext, const std::string & rLocation)
{
	m_ContCat.cLogMessage.RequestControl(CInstrumentCatalog::sLogMessage(rLevel, "MFCManager.Controller"+ rContext, rMessage, rLocation));
}

void CMFCController::l_SendErrorMessage(const CMFCDevice::sMFCDeviceError & rErrorMessage)
{
	MTL_Assert(CMFCDevice::sMFCDeviceError::kNoError != rErrorMessage.Level);
	if(CMFCDevice::sMFCDeviceError::kNoError != rErrorMessage.Level)
		m_ContCat.cError.RequestControl(rErrorMessage);
}

void CMFCController::l_SendSearchProgress(const U8 & rSearchProgress)
{
	m_ContCat.cSearchProgress.RequestControl(rSearchProgress);
}

void CMFCController::l_SendSearchResult(const CInstrumentCatalog::sSearchResult & rSearchResult)
{
	m_ContCat.cSearchResult.RequestControl(rSearchResult);
}

void CMFCController::l_SendMeasResult(const CInstrumentCatalog::sMeasurementResults & rMeasRes)
{
	m_ContCat.cMeasResults.RequestControl(rMeasRes);
}

bool CMFCController::l_IsAbortRequested()
{
	return CThreadedMailBox::l_IsRequestedToStop();
}

void CMFCController::l_CheckConnectionOnWarning(const CMFCDevice::sMFCDeviceError & rOriginError, CMFCDevice::sMFCDeviceError & rErrorOut)
{
	rErrorOut = rOriginError;
	if (CMFCDevice::sMFCDeviceError::eLevel::kWarning == rOriginError.Level && nullptr != m_pMFCDevice)
	{
		if(!m_pMFCDevice->IsInstrumentConnected())
		{
			rErrorOut.Level = CMFCDevice::sMFCDeviceError::eLevel::kError;
			rErrorOut.Description = "No instrument connected";
			rErrorOut.Location = rErrorOut.Location;
		}
	}
}

void CMFCController::l_AbortInstrument()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	try
	{
		if (nullptr != m_pMFCDevice)
		{
			if (!m_pMFCDevice->AbortInstrument())
			{
				if (!m_pMFCDevice->AbortInstrument())
					throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kError, "Could not abort instrument", MTL__LOCATION__);
			}
		}
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		// Signal error
		l_SendErrorMessage(rE);
	}
}

void CMFCController::l_ScanResources()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	try
	{
		// Check initial conditions
		MTL_Assert(nullptr != m_pRsrcManager);
		if (nullptr == m_pRsrcManager)
			throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "No given resource manager", MTL__LOCATION__);

		static CParsedResourceList l_RsrcList, l_NewRsrcList;
		// Update resource list
		if (!m_pRsrcManager->FindResources(l_NewRsrcList, "?*"))
			throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kError, "Could not look for resources : " + MTL::Instrument::StatusDescription(m_pRsrcManager->Session(), m_pRsrcManager->Status()), MTL__LOCATION__);
		if (l_RsrcList != l_NewRsrcList)
		{
			l_RsrcList = l_NewRsrcList;
			m_ContCat.cInstrumentList.RequestControl(l_NewRsrcList);
		}
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		// Signal error
		l_SendErrorMessage(rE);
		// Send empty list
		m_ContCat.cInstrumentList.RequestControl(CParsedResourceList());
	}
}

void CMFCController::l_GetRemoteBoxButtonStatus()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	try {
		CMFCDevice::sMFCDeviceError l_RetError;
		bool l_RemoteBoxPressed(false);
		// Check initial conditions
		MTL_Assert(nullptr != m_pMFCDevice);
		if (nullptr == m_pMFCDevice)
			throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given instrument pointer is null", MTL__LOCATION__);

		l_RetError = m_pMFCDevice->RemoteBoxButtonStatus(l_RemoteBoxPressed);
		if(l_RetError.HasError())
		{
			throw l_RetError;
		}
		if (l_RemoteBoxPressed)
			m_ContCat.cRemoteBoxStart.RequestControl(true);
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		CMFCDevice::sMFCDeviceError l_Error;
		l_CheckConnectionOnWarning(rE, l_Error);
		// Signal error
		l_SendErrorMessage(l_Error);
	}
}

bool CMFCController::l_ScanPA()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	bool l_PAInfoChanged(false);
	bool l_PASucceeded(true);
	CInstrumentCatalog::sInstrumentConfiguration l_InsConf;
	try
	{
		CMFCDevice::sMFCDeviceError l_RetError;

		// Check initial conditions
		MTL_Assert(nullptr != m_pMFCDevice);
		if (nullptr == m_pMFCDevice)
		{
			if (!m_PAInfo.ProbeList.empty())
			{
				// Clear whole PA Info including FCA information
				m_PAInfo = CInstrumentCatalog::sProbeArrayInformation();
				l_InsConf = CInstrumentCatalog::sInstrumentConfiguration();
				l_PAInfoChanged = true;
			}
			throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given instrument pointer is null", MTL__LOCATION__);
		}

		l_RetError = m_pMFCDevice->ScanPA(m_PAInfo, l_InsConf, l_PAInfoChanged);
		if(l_RetError.HasError())
		{
			throw l_RetError;
		}
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		CMFCDevice::sMFCDeviceError l_Error;
		l_CheckConnectionOnWarning(rE, l_Error);
		// Signal error
		l_SendErrorMessage(l_Error);
		l_PASucceeded = false;
	}
	if (l_PAInfoChanged) // Send new PA Info if changed
	{
		m_ContCat.cPAInfo.RequestControl(m_PAInfo);
		m_ContCat.cInstrumentConfiguration.rRequestControl(l_InsConf);
	}
	return l_PASucceeded;
}

void CMFCController::l_OperateWithInstrument()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	try
	{
		CMFCDevice::sMFCDeviceError l_RetError;
		// Depending on the requested operating mode
		switch (m_InstrState.OperatingMode)
		{
		//------------------------------------------------
		// Idle operation
		case CInstrumentCatalog::eOperatingMode::kIdle:
			// Nothing to do
			throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Cannot operate in kIdle mode", MTL__LOCATION__);
			break;

			//------------------------------------------------
			// Search operation
		case CInstrumentCatalog::eOperatingMode::kSearch:
		{
				// Check initial conditions
				// Check that we have an instrument
				MTL_Assert(nullptr != m_pMFCDevice);
				if (nullptr == m_pMFCDevice)
					throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given instrument pointer is null", MTL__LOCATION__);
				l_RetError = m_pMFCDevice->Search(m_InstrState);
				if(l_RetError.HasError())
				{
					throw l_RetError;
				}
		}
			break;

			//------------------------------------------------
			// Measure operation
		case CInstrumentCatalog::eOperatingMode::kMeasure:
		{
				// Check initial conditions
				// Check that we have an instrument
				MTL_Assert(nullptr != m_pMFCDevice);
				if (nullptr == m_pMFCDevice)
					throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given instrument pointer is null", MTL__LOCATION__);
				l_RetError = m_pMFCDevice->Measure(m_InstrState);
				if(l_RetError.HasError())
				{
					throw l_RetError;
				}
		}
			break;
		}
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		// Signal error
		l_SendErrorMessage(rE);
	}
}

void CMFCController::l_RequestSelfAbort()
{
	Terminate();

	m_ControllerState = CMFCManagerCatalog::eControllerState::kIdle;
	m_ContCat.cState.RequestControl(m_ControllerState);
}

CMFCController::CMFCController(CMFCManagerCatalog & rCatInstance, const std::string & rDefaultPAConfScriptFilePath, const std::string & rCustomPAConfScriptDir)
	:
	  m_ContCat(rCatInstance),
	  m_Timer(std::bind(&CMFCController::l_PostTimerEvent, this)),
	  m_ControllerState(CMFCManagerCatalog::eControllerState::kIdle),
	  m_pMFCDevice(nullptr),
	  m_DefaultPAConfScriptFilePath(rDefaultPAConfScriptFilePath),
	  m_CustomPAConfScriptDirectory(rCustomPAConfScriptDir)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
}

CMFCController::~CMFCController()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	Terminate();
}

void CMFCController::l_TimerEventHandler()
{
	try
	{
		switch (m_ControllerState)
		{
		case CMFCManagerCatalog::eControllerState::kScanningInstruments:
			l_ScanResources();
			m_Timer.Stop();
			m_Timer.StartSingleShot(std::chrono::milliseconds(RSRC_POLLING_PERIOD_MS));
			break;

		case CMFCManagerCatalog::eControllerState::kScanningPA:
			if(l_ScanPA())
				l_GetRemoteBoxButtonStatus();
			m_Timer.Stop();
			m_Timer.StartSingleShot(std::chrono::milliseconds(PA_POLLING_PERIOD_MS));
			break;

		case CMFCManagerCatalog::eControllerState::kSearching:
		case CMFCManagerCatalog::eControllerState::kMeasuring:
			l_OperateWithInstrument();
			Terminate();
			break;

		default:
			// Unhandled case
			MTL_Assert(false);
			throw false;
			break;
		}
	}
	catch (...)
	{
		l_SendErrorMessage(CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Unhandled exception type", MTL__LOCATION__));
		Terminate();
	}
}

void CMFCController::l_PostTimerEvent()
{
	CThreadedMailBox::l_SendMsg(sMFCControllerMailBoxMsg(sMFCControllerMailBoxMsg::eMsgType::kTimerEvent));
}

void CMFCController::l_ThreadTask(eEventType EvtType, sMFCControllerMailBoxMsg * pMsg)
{
	switch (EvtType)
	{
	case eEventType::kExit:
		// Abort instrument
		l_AbortInstrument();
		// Set new controller state
		m_ControllerState = CMFCManagerCatalog::eControllerState::kIdle;
		m_ContCat.cState.RequestControl(m_ControllerState);
		break;
	case eEventType::kUserMsg:
		if (nullptr != pMsg)
		{
			switch (pMsg->MsgType)
			{
			case sMFCControllerMailBoxMsg::eMsgType::kTimerEvent:
				l_TimerEventHandler();
				break;
			}
		}
		break;
	}
}

void CMFCController::Operate()
{
	CThreadedMailBox::ThreadStart();
}

void CMFCController::Terminate()
{
	CThreadedMailBox::ThreadStop();
	m_Timer.Stop(); // Must be done after CThreadedMailBox::ThreadStop(). Timer may have been restarted before the thread stop completely.
}

void CMFCController::ScanInstruments(CVISAResourceManager * pRsrcManager)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Prepare context for scanning instruments
	// Start timer to trigger periodic scan

	Terminate();

	m_pMFCDevice = nullptr;
	MTL_Assert(nullptr != pRsrcManager);
	if (nullptr == pRsrcManager)
	{
		l_SendErrorMessage(CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given resource manager pointer is null", MTL__LOCATION__));
		return;
	}
	m_pRsrcManager = pRsrcManager;

	m_ControllerState = CMFCManagerCatalog::eControllerState::kScanningInstruments;
	m_ContCat.cState.RequestControl(m_ControllerState);
	Operate();
	m_Timer.StartSingleShot(std::chrono::milliseconds(0));
}

void CMFCController::ScanPA(CMFCDevice * pMFCDevice)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Prepare context for scanning PA
	// Start timer to trigger periodic scan

	Terminate();

	MTL_Assert(nullptr != pMFCDevice);
	if (nullptr == pMFCDevice)
	{
		l_SendErrorMessage(CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given instrument pointer is null", MTL__LOCATION__));
		return;
	}
	if (pMFCDevice != m_pMFCDevice)
	{
		m_PAInfo.clear(); // Clear PA info
	}
	m_pMFCDevice = pMFCDevice;

	m_ControllerState = CMFCManagerCatalog::eControllerState::kScanningPA;
	m_ContCat.cState.RequestControl(m_ControllerState);
	Operate();
	m_Timer.StartSingleShot(std::chrono::milliseconds(0));
}

void CMFCController::SetInstrumentState(CMFCDevice * pMFCDevice, CInstrumentCatalog::sInstrumentState InsState)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Prepare context for changing instrument state
	// If requested to go Idle, nothing else is done
	// If requested to operate, trigger operation with a single timer shot

	Terminate();

	MTL_Assert(nullptr != pMFCDevice);
	if (nullptr == pMFCDevice)
	{
		l_SendErrorMessage(CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given instrument pointer is null", MTL__LOCATION__));
		return;
	}
	m_pMFCDevice = pMFCDevice;
	m_InstrState = InsState;

	switch (m_InstrState.OperatingMode)
	{
	case CInstrumentCatalog::eOperatingMode::kIdle:
		m_ControllerState = CMFCManagerCatalog::eControllerState::kIdle;
		m_ContCat.cState.RequestControl(m_ControllerState);
		// Nothing to do
		break;
	case CInstrumentCatalog::eOperatingMode::kSearch:
		m_ControllerState = CMFCManagerCatalog::eControllerState::kSearching;
		m_ContCat.cState.RequestControl(m_ControllerState);
		Operate();
		m_Timer.StartSingleShot(std::chrono::milliseconds(0));
		break;
	case CInstrumentCatalog::eOperatingMode::kMeasure:
		m_ControllerState = CMFCManagerCatalog::eControllerState::kMeasuring;
		m_ContCat.cState.RequestControl(m_ControllerState);
		Operate();
		m_Timer.StartSingleShot(std::chrono::milliseconds(0));
		break;
	}
}

void CMFCController::UpdateNormalizationDate()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	try {
		MTL_Assert(nullptr != m_pMFCDevice);
		if (nullptr == m_pMFCDevice)
			throw CMFCDevice::sMFCDeviceError(CMFCDevice::sMFCDeviceError::eLevel::kFatal, "Given instrument pointer is null", MTL__LOCATION__);

		CMFCDevice::sMFCDeviceError l_RetError;
		CInstrumentCatalog::tISO8601DateTime l_NewNormDate;
		l_RetError = m_pMFCDevice->GetNormalizationDate(l_NewNormDate);
		if(l_RetError.HasError())
		{
			throw l_RetError;
		}
		if(l_NewNormDate != m_PAInfo.NormalizationDate)
		{
			m_PAInfo.NormalizationDate = l_NewNormDate;
			m_ContCat.cPAInfo.RequestControl(m_PAInfo);
		}
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		CMFCDevice::sMFCDeviceError l_Error;
		l_CheckConnectionOnWarning(rE, l_Error);
		// Signal error
		l_SendErrorMessage(l_Error);
	}
}
