//////////////////////////////////////////////////////////////////////////
/// \file		MFC3045.h
/// \brief 		Interface definition for C++ API for Metrolab MFC3045.
/// \date       2020
/// \copyright 	Metrolab Technology SA
/// \author 	S.Benz


#pragma once

// Standard includes
#include <ctime>

// Personal includes
#include "VISAInstrument.h"
#include "MFC3045Types.h"

// Version
#define MFC3045_CPP_DRIVER_BCD_VERSION		0x0100


using namespace MTL::Instrument::MFC3045Types;

namespace MTL {
	namespace Instrument {
		
		/// \brief MFC3045 instrument class.
		class CMFC3045Instrument : protected CVISAInstrument
		{
			//------------------------------------------------------------------------//
			//						Private Attributes
			//------------------------------------------------------------------------//
		private:
			//General
			U32						m_WaitTimeRdData_ms;	// time used in l_WaitAvailableDataRX()
			bool					m_AbortSerialTransfert;
			CVISABuffer				m_ReadBuffer;	
			sSerialPortSettings		m_HostSerialConfig;		// store current VISA Serial settings. Must only be updated in ConnectInstrument()
			eAdvancedMode			m_AdvancedMode;			// updated by Read/Write of register ADV
			eDataBlockMode			m_DataBlockMode;		// updated by Read/Write of register BLK
			U8						m_NbProbesInPA;			// updated by Read/Write of register NPR
			bool					m_SMAEnabled;
			uSMAByte				m_SMAValue;
			
			//------------------------------------------------------------------------//
			//						Private Utilities
			//------------------------------------------------------------------------//
		private:
			// Serial port configuration
			bool l_SerialPortHostConfig(const sSerialPortSettings & rSerialSettings);
			// Serial buffer utilities
			bool l_ClearHostSerialBufferRxTx();
			bool l_GetNewDataRxBufferLength(U32 & rRxBufferLength);
			bool l_WaitAvailableDataRX(const U32 & rTimeoutRx, U32 & rRXBufferLength, const bool & rErrorOnTimoute=true);
			bool l_ExtractSmaMsgFromDataRX(const std::string & rNewRxBufferStrIn, uSMAByte & rSMAValue, std::string & rNewRxBufferStrOut);
			// Serial write and read
			bool l_Write(const std::string & rWriteStr, const bool & rForceWrite=false);
			bool l_ReadString(std::string & rRxBufferStr, const U8 & rNStringAssembly=1, const bool & rCheckSmaMsg=false);
			bool l_ReadNCharNoTerm(CVISABuffer & rReadBuffer, const U16 & rReadNChar);

			// Date conversions
			bool l_DateStrToDateTimeUtc(const std::string & rUtcDateStr, std::time_t & rUtcDateTime);
			bool l_DateTimeToDateStrUtc(const std::time_t & rUtcDateTime, std::string & rUtcDateStr);

			//------------------------------------------------------------------------//
			//						MFC3045 Interface
			//------------------------------------------------------------------------//
		public:
			///
			/// \brief Constructors / destructors
			///
			//------------------------------------------------------------------------//
			// Constructors / Destructors
			CMFC3045Instrument(CVISAResourceManager & rRM, tResourceName Rsrc);
			virtual ~CMFC3045Instrument();

			//------------------------------------------------------------------------//
			/// @name Open / Close, Initialized the serial Host communication and test connection to instrument.
			///
			/// Create a link to the instrument. The actual handle to the instrument is private to the class.
			///	
			///@{
			// Connect, check connection, disconnect to Instrument with Serial Port (VISA)
			bool CheckInstrumentConnection(const U8 & rl_Retry = 3);
			bool ConnectInstrument(const sSerialPortSettings & rSerialSettings, const U32 & rInitialTimeout_ms = RS_RDWR_VISA_TIMEOUT_MS);
			bool DisconnectInstrument();
			///@}

			//------------------------------------------------------------------------//			
			/// @name special public utilities
			///
			///	Abort or allow serial command
			///	Read a custom command
			/// Set maximum time in ms to wait before expecting new data in RX buffer to be read
			///@{
			// Special public utilities
			bool Abort();	// abort any serial command in process and send a BRK cmd (break any acquisition) to instrument.
			bool ReadWithCustomCmd(const std::string & rCommandStr, std::string & rRegValStr);
			void GetCurrentTimeoutRdDataMs(U32 & rRdDataTimeout_ms);
			void SetTimeoutRdNewDataMs(const U32 & rNewdataTimeout_ms = RS_WAIT_DATARD_MS); // set time to wait before to read any data (set m_WaitTimeRdData_ms for l_WaitAvailableDataRX() use)
			///@}

			//------------------------------------------------------------------------//			
			/// @name Main unit RS232 settings
			///
			///  Accessors to get or set the serial RS232 parameters in the MFC main unit  (Advanced mode, System Reset, Versions, Serial number, Remote Led, ...)
			///@{
			// RS232 setting register (RSP)
			bool ParseParmForRSP(const sSerialPortSettings & rSerialSettings, U16 & rParmRSPSettings);
			bool ParseParmFromRSP(const U16 & rParmRSPSettings, sSerialPortSettings & rSerialSettings);
			bool SerialRSPSet(const U16 & rRegVal);
			bool SerialRSPGet(U16 & rRegVal);
			///@}

			//------------------------------------------------------------------------//			
			/// @name Miscellaneous MFC commands
			///
			///  Accessors to get or set some MFC  (Advanced mode, System Reset, Versions, Serial number, Remote Led, ...)
			///@{
			// Miscellaneous Commands
				//Advance command mode register (ADV)
			bool AdvancedCommandModeSet(const eAdvancedMode & rRegVal);
			bool AdvancedCommandModeGet(eAdvancedMode & rRegVal);
				// Reset register (RST)
			bool SystemResetSet();
				// Version register (VER)
			bool VersionGet(std::string & rVersionInfo, F32 & rFwVersion);
			bool VersionGet(std::string & rVersionInfo, F32 & rFwVersion, const eWhoVersion & rSelect);
				// Serial Number register (S/N)
			bool SerialNumberGet(std::string & rSerialNum);
			bool SerialNumberGet(std::string & rSerialNum, const eWhoSerialNum & rSelect);
				// Remote Box Busy Led register (LED)
			bool RemoteBusyLedSet(const eRemoteBusyLed & rRemoteLed);
			bool RemoteBusyLedGet(eRemoteBusyLed & rRemoteLed);
				// DDS Fixed Frequency register (DFF)
			bool DdsFrequencySet(const U32 & rRegVal);
			bool DdsFrequencyGet(U32 & rRegVal);
				// Frequency Sweep register (FSW)
			bool FrequencySweepSet(const eSweepFreqMode & rSweepMode, const U32 & rFreqStepVal =0);
				// Frequency Sweep shift
			bool FrequencySweepShiftSet(const eShiftFreq & rShiftDirection);
			///@}

			//------------------------------------------------------------------------//			
			/// @name Status registers
			///
			///  Accessors to get the MFC3045 status registers value (error, information, ...).  
			///@{			
			// Status Handling and command Error
			bool StatusGet(const eStatusRegister & rStatRegSel, U8 & rStatus);
			bool StatusGet(const std::vector<eStatusRegister> & rStatRegSel, std::vector<U8> & rStatus);
			bool LastCmdErrorGet(std::string & rCmdError);
			bool SMAGet(U8 & rRegVal);
			bool DisableSMA();
			///@}
		private:
			bool l_SMASet(const U8 & rRegVal);


			//------------------------------------------------------------------------//			
			/// @name Modulation parameters
			///
			///  Accessors to configure the way the RF modulation is processed to make measurement.  
			///@{			
			// Modulation Parameters
				// Modulation Amplitude ppm Register (MDA)
		public:
			bool ModulationAmplitudeSet(const U32 & rRegVal);
			bool ModulationAmplitudeGet(U32 & rRegVal);
				// Modulation Central Frequency dHz Register (MCF)
			bool ModulationCentralFreqSet(const U32 & rRegVal);
			bool ModulationCentralFreqGet(U32 & rRegVal);
				// Modulation Lowest Frequency dHz Register (MLF)
			bool ModulationLowestFreqSet(const U32 & rRegVal);
			bool ModulationLowestFreqGet(U32 & rRegVal);
				// Modulation Highest Frequency dHz Register (MHF)
			bool ModulationHighestFreqSet(const U32 & rRegVal);
			bool ModulationHighestFreqGet(U32 & rRegVal);
				// Modulation Reference Register (MRE)
			bool ModulationReferenceSet(const eModulationRef & rRegVal);
			bool ModulationReferenceGet(eModulationRef & rRegVal);
				// Modulation Period ms Register (MDP)
			bool ModulationPeriodSet(const U16 & rRegVal);
			bool ModulationPeriodGet(U16 & rRegVal);
				// Number of Step for Ramp Modulation Register (NSR)
			bool ModulationNumberStepRampGet(U32 & rRegVal);
				// Number of Step for Plateau Modulation Register (NSP)
			bool ModulationNumberStepPlateauGet(U32 & rRegVal);
				// DDS Bit Resolution Register (DBR)
			bool DdsBitResolutionGet(U32 & rRegVal);
			bool DdsBitResolutionFdhzGet(F32 & rDdsResolutionFdhz);
			///@}

			//------------------------------------------------------------------------//			
			/// @name Measurement parameters
			///
			///  Accessors to configure the way the measurement is processed to compute the result.  
			///@{
			// Measurement Parameters
				// Number of Measurement Cycles Register (NCY)
			bool NumberMeasCycleSet(const U16 & rRegVal);
			bool NumberMeasCycleGet(U16 & rRegVal);
				// Number of Preliminary Cycles Register (NPC)
			bool NumberPreCycleSet(const U16 & rRegVal);
			bool NumberPreCycleGet(U16 & rRegVal);
				// Time duration of Preliminary Cycles Register (NPT)
			bool TimePreCycleSet(const U16 & rRegVal);
			bool TimePreCycleGet(U16 & rRegVal);
				// Rejection Signal Offset Register (RSO)
			bool RejectionOffsetSet(const U16 & rRegVal);
			bool RejectionOffsetGet(U16 & rRegVal);
				// Rejection Signal Gap Register (RSG)
			bool RejectionGapSet(const U16 & rRegVal);
			bool RejectionGapGet(U16 & rRegVal);
				// Time Versus Precision Register (TVP)
			bool TimeVsPrecisionSet(const ePrecision & rRegVal);
			bool TimeVsPrecisionGet(ePrecision & rRegVal);
			///@}

			//------------------------------------------------------------------------//
			/// @name Measurement operation
			///
			///  Accessors to control and configure the measurement operating mode (Start, Stop, Search, Continuous measurement)
			///@{
			//------------------------------------------------------------------------//
			// Measurement operating mode 
				// Starts a Measurement of (NCY Cycles), Register (RUN)
			bool MeasCycleStart();
			bool MeasCycleStart(const U8 & rProbeSel);
			bool MeasStartWithSMA();
			bool MeasStartWithSMA(const U8 & rProbeSel);
			bool WaitSmaMeasureDataReady(const U32 & rMeasWaitTimeout_ms, uSMAByte & rDataReady);

				// Search NMR over full range PA, Register (SRC)
			bool SearchStart();
			bool SearchStart(const U8 & rProbeSel);
			bool SearchStartWithSMA();
			bool SearchStartWithSMA(const U8 & rProbeSel);
				// Starts Continuous Measurement, Register (CTN)
			bool ContinuousMeasStart();
				// Break: Interrupts (Stop) data measurement acquisition, Register (BRK)
			bool BreakMeasSet();
			///@}

			//------------------------------------------------------------------------//		
			/// @name Measurement data reading
			///
			///  Accessors to get measurement data results (NMR, Standard deviation, Valid cycle, statistics)
			///@{
			// Data Reading
		private:
			bool l_ReadEndTransmissionBlockChar();
			bool l_ReadMeasurementTableBlock(const std::string rStringCMD, std::vector<U32> & rData);
				// Data transfert Block Mode Register (BLK)		 
			bool l_DataBlockModeSet(const eDataBlockMode & rRegVal); // we force the Block mode to decimal Block for the whole API
		public:
			bool DataBlockModeDecimalSet();
			bool DataBlockModeGet(eDataBlockMode & rRegVal);

				// Magnetic Field Value dHz Register (BFV)
			bool DataMagneticFieldGet(const U8 & rNbMeasurements, std::vector<U32> & rData);
			bool DataMagneticFieldGet(U32 & rData, const U8 & rProbeSel);
				// Standard Deviation dHz Register (BSD)
			bool DataStdDeviationGet(const U8 & rNbMeasurements, std::vector<U32> & rData);
			bool DataStdDeviationGet(U32 & rData, const U8 & rProbeSel);
				// Number of Cycles Register (BNC)
			bool DataNumberValidCycleGet(const U8 & rNbMeasurements, std::vector<U32> & rData);
			bool DataNumberValidCycleGet(U32 & rData, const U8 & rProbeSel);
				// Individual NMR Frequencies (up & down) dHz Register (BIN)
			bool DataIndividualFrequencyGet(U32  & rDataUp, U32 & rDataDown);
				// Central NMR Frequency dHz Register (BFC)
			bool DataCentralFrequencyGet(U32 & rData);
				// Lowest NMR Frequency dHz Register (BFL)
			bool DataLowestFrequencyGet(U32 & rData);
				// Highest NMR Frequency dHz Register (BFH)
			bool DataHighestFrequencyGet(U32 & rData);
				// Difference Highest-Lowest NMR Frequency ppm Register (BFD)
			bool DataDifferenceFrequencyGet(U32 & rData);
			///@}

			//------------------------------------------------------------------------//			
			/// @name Probe Array information
			///
			///  Accessors to get probe array information (Number of probes, frequency range, ...)
			///@{
			// Probe Array Information
				// Number of PRobes (measurement points) register (NPR)
			bool NumberProbesSet(const U8 & rRegVal);
			bool NumberProbesGet(U8 & rRegVal);
				// Probe Array Central Frequency register (PCF)
			bool PaCentralFreqSet(const U32 & rRegVal);
			bool PaCentralFreqGet(U32 & rRegVal);
				// Probe Array Lowest Frequency register (PLF)
			bool PaLowestFreqSet(const U32 & rRegVal);
			bool PaLowestFreqGet(U32 & rRegVal);
				// Probe Array Highest Frequency register (PHF)
			bool PaHighestFreqSet(const U32 & rRegVal);
			bool PaHighestFreqGet(U32 & rRegVal);
				// RF Harmonic Register (RFH)
			bool RfHarmonicSet(const eRFH & rRegVal);
			bool RfHarmonicGet(eRFH & rRegVal);
				// NMR Signal Threshold register (NST)
			bool NmrThresholdSet(const U8 & rRegVal);
			bool NmrThresholdGet(U8 & rRegVal);
			///@}

			//------------------------------------------------------------------------//
			/// @name Main unit and Probe Array EEPROMs access
			///
			///  Accessors to set or get MFC main unit and Probe Array EEPROM contente.
			///@{
			// Main unit EEPROM, Probe Array EEPROM and Probe Array Calibration (Normalization)
				// Settings in Main Unit EEPROM register (SMU)
			bool MainUnitEepromWrite(); // write dedicated RAM content to MU EEPROM
			bool MainUnitEepromRead(std::vector<char> & rEEPROMContent);
				// Settings in Probe Array EEPROM register (SPA)
			bool ProbeArrayEepromWrite();// write dedicated RAM content to PA EEPROM
			bool ProbeArrayEepromRead(std::vector<char> & rEEPROMContent);
			///@}

			//------------------------------------------------------------------------//
			/// @name Probe Array Normalization
			///
			///  Accessors to control the normalization of the Probe Array.
			///@{
				// Calibration Probe measurement table Selection register (CPS) ; (Measurement Table without correction. Stored in RAM)
			bool CalibrationProbeMeasurementSet(const U8 & rProbeSel);
			bool CalibrationProbeMeasurementTableClear();
			bool CalibrationProbeMeasurementTableGet(const U8 & rNbMeasurements, std::vector<U32> & rData);
			bool CalibrationMeasurementAverageProbeGet(U32 & rData);
				// Calibration Build correction Table register (CBT); (Correction Table in RAM)
			bool CalibrationBuildTableCopyEEPROMToRAM();
			bool CalibrationBuildTableClearRAM();
			bool CalibrationBuildTableWithTargetSet(const U32 & rTarget);
			bool CalibrationBuildTableGet(const U8 & rNbMeasurements, std::vector<I32> & rData);
			bool CalibrationBuildProbeGet(I32 & rData, const U8 & rProbeSel);
				// Calibration Edit correction Table register (CET)
			bool CalibrationEditTableSet(const I32 & rData);
			bool CalibrationEditTableSet(const I32 & rData, const U8 & rProbeSel);
			bool CalibrationEditTableGetProbePointed(U8 & rProbeIdxCET);
			
				// Calibration Date of Probe Array register (CDP)
			bool CalibrationDatePASet(const U8 & rDayDD, const U8 & rMonthMM, const U8 & rYearYY, const bool & rSaveInEeprom = false);
			bool CalibrationDatePASet(const std::time_t & rUtcDate, const bool & rSaveInEeprom = false);
			bool CalibrationDatePAGet(std::string & rStrDate, std::time_t & rUtcDate);
				// Calibration Date of Main Unit register (CDU)
			bool CalibrationDateMUSet(const U8 & rDayDD, const U8 & rMonthMM, const U8 & rYearYY, const bool & rSaveInEeprom = false);
			bool CalibrationDateMUSet(const std::time_t & rUtcDate, const bool & rSaveInEeprom = false);
			bool CalibrationDateMUGet(std::string & rStrDate, std::time_t & rUtcDate);
			///@}
		};

	}
}
