#pragma once

// Standard includes
#include <string>
#include <vector>
#include <cmath>

// Personnal includes
#include "OSDefines.h"

// Personal Constant defines
// registers limit values
#define MFC_ASCII							"MFC"
#define MFC_ASCII_DEC						777067
#define RSP_MIN								0
#define RSP_MAX								1855
#define FSW_DFF_MAX							440000000
#define MDA_MIN								200
#define MDA_MAX								16777216	//((unsigned long)(2^24))
#define MOD_FREQ_MIN_DHZ					10000000	//((unsigned long)(10^7))
#define MOD_FREQ_MAX_DHZ					3080000000	//((unsigned long)(308*10^7))
#define PA_FREQ_MIN_DHZ						2000000		//((unsigned long)(2*10^6))
#define PA_FREQ_MAX_DHZ						3080000000	//((unsigned long)(308*10^7))
#define NCY_MIN								2
#define NCY_MAX								1500
#define NPC_MAX								100
#define NPT_MAX								30000
#define RSO_MIN								1
#define RSO_MAX								20000
#define RSG_MIN								1
#define RSG_MAX								1000
#define NPR_MAX								96
#define CORRECT_POBE_MIN_DHZ				-37768
#define CORRECT_POBE_MAX_DHZ				37767

//Serial miscellaneous constants
#define DATA_END_ETB						23
#define RS_RD_BUFF_DEFAULT_SIZE				2048
#define RS_RDWR_VISA_TIMEOUT_MS				5000
#define RS_WAIT_DATARD_MS					5000
#define PA_EEPROM_CHARSIZE					576
#define MU_EEPROM_CHARSIZE					160
#define CHECKSUM_CHARSIZE					4
#define PA_CALIB_TABLE_SIZE					96
#define READ_EEPROM_POLL_PERIOD_MS			10
#define MSG_EEPROM_ERASED					"Flash PROM erased"
#define FW_MIN_VERSION						2.0

#define GYRORATIO_MFC3048_HZT				42576255 // HZ/T

namespace MTL {
	namespace Instrument {
		namespace MFC3045Types {
			//----------------------------------------------------------------------//
			//						Modulation types								//
			//----------------------------------------------------------------------//
			// Modulation Reference
			enum class eModulationRef : U8 {
				kMDA = 0,
				kMCF,
				kMLF,
				kMHF
			};

			//----------------------------------------------------------------------//
			//						Measurement types								//
			//----------------------------------------------------------------------//
			// Time versus Precision
			enum class ePrecision : U8 {
				kHighPrecision = 0,	// Maximum Precision, Longer time
				kLowPrecision = 1	// Lower Precision, Shorter time
			};

			//----------------------------------------------------------------------//
			//						Data Reading types								//
			//----------------------------------------------------------------------//
			//Data transfert Block Mode
			enum class eDataBlockMode : U8 {
				kOneByOne = 0,
				kAllDecimal,
				kAllHexa
			};

			//----------------------------------------------------------------------//
			//							Status	types								//
			//----------------------------------------------------------------------//
			enum class eStatusRegister : U8 {
				kStatus1 = 0,
				kStatus2,
				kStatus3,
				kStatus4,
				kStatus5,
				kStatus6
			};

			// Status register 1 (Main Instrument Event and Error); Read only
			union uStatusByte1 {
				U8	RawSTB1 = 0;
				struct sStatusByte1 {
					// Contains a 1-byte status register 1. The MFC3045 uses the following bits:
					U8		DRDY : 1;	// bit[0]: Data Ready
					U8		ECMD : 1;	// bit[1]: Command Error
					U8		EMOD : 1;	// bit[2]: Modulation Error
					U8		ECOM : 1;	// bit[3]: Communication Error
					U8		RBCS : 1;	// bit[4]: Remote button change status
					U8		EMIS : 1;	// bit[5]: Miscellaneous Error
					U8		NNMR : 1;	// bit[6]: No NMR Signal on Probe(s)
					U8		PWON : 1;	// bit[7]: Power on (This bit is set to 1 at power on or following a reset of the instrument.)
				} StatusByte1;
			};

			// Status register 2 (Modulation parameters errors)
			union uStatusByte2 {
				U8 RawSTB2 = 0;
				struct sStatusByte2 {
					// Contains a 1-byte status register 2. The MFC3045 uses the following bits:
					U8		PSS : 1;	// bit[0]: Period shorter than setting
					U8		PLS : 1;	// bit[1]: Period longer than setting
					U8		MAR : 1;	// bit[2]: Max amplitude reach
					U8		PNS : 1;	// bit[3]: Period not stable
					U8		CYL : 1;	// bit[4]: Cycle lost
					U8		LCR : 1;	// bit[5]: Allowed lost cycle reached
					U8		MED : 1;	// bit[6]: Modulation parameter exceeded DDS Range
					U8		PANC : 1;	// bit[7]: Probe-Array not connected
				} StatusByte2;
			};

			// Status register 3 (Measurement status)
			union uStatusByte3 {
				U8 RawSTB3 = 0;
				struct sStatusByte3 {
					// Contains a 1-byte status register 3. The MFC3045 uses the following bits:
					U8		DAA : 1;	// bit[0]: Data available
					U8		MRA : 1;	// bit[1]: Measurement (Run) active
					U8		MCA : 1;	// bit[2]: Continuous measurement active
					U8		SEA : 1;	// bit[3]: Search active
					U8		NU1 : 1;	// bit[4]: Not Used
					U8		RFE : 1;	// bit[5]: RF generator Enable (ON)
					U8		REL : 2;	// bit[6-7]: Remote Led
										// 00 Led off
										// 01 Led is Lit
										// 10 Led is blinking slowly
										// 11 Led is blinking quickly
				} StatusByte3;
			};

			// Status register 4 (Miscellaneous Status)
			union uStatusByte4 {
				U8 RawSTB4 = 0;
				struct sStatusByte4 {
					// Contains a 1-byte status register 4. The MFC3045 uses the following bits:
					U8		HWP : 1;	// bit[0]: Hardware Problem
					U8		DCR : 1;	// bit[1]: Data Copied in RAM from p-a EEPROM
					U8		OCT : 1;	// bit[2]: Overflow in creating Calibration Table
					U8		RBS : 1;	// bit[3]: Remote Button State
					U8		RBC : 1;	// bit[4]: Remote Button Change state
					U8		MSW : 1;	// bit[5]: Memory (EEPROM) successfully wrote
					U8		EMM : 1;	// bit[6]: Error with Main unit Memory (EEPROM)
					U8		EPM : 1;	// bit[7]: Error with Probe-Array Memory (EEPROM)
				} StatusByte4;
			};

			// Status register 5 (Baud Rate & Communication error)
			union uStatusByte5 {
				U8 RawSTB5 = 0;
				struct sStatusByte5 {
					// Contains a 1-byte status register 5. The MFC3045 uses the following bits:
					U8		RBR : 3;	// bit[0-2]: RS-232 Baud Rate
										//	0 0 0 2400
										//	0 0 1 4800
										//	0 1 0 9600  (Default value)
										//	0 1 1 19200
										//	1 0 0 28800
										//	1 0 1 38400
										//	1 1 0 57600
										//	1 1 1 115200
					U8		ECS : 1;	// bit[3]: Error Check Sum
					U8		EOR : 1;	// bit[4]: Error OverRun
					U8		EPY : 1;	// bit[5]: Error Parity
					U8		EFR : 1;	// bit[6]: Error Framing
					U8		NU2 : 1;	// bit[7]: Not used
				} StatusByte5;
			};

			// Status register 6 (Communication settings)
			union uStatusByte6 {
				U8 RawSTB6 = 0;
				struct sStatusByte6 {
					// Contains a 1-byte status register 6. The MFC3045 uses the following bits:
					U8		BL8 : 1;	// bit[0]: Bit length is 8 (set), 7 (cleared)
					U8		SB2 : 1;	// bit[1]: Stop bit is 2 (set), 1 (cleared)
					U8		PAS : 1;	// bit[2]: Parity set
					U8		PAE : 1;	// bit[3]: Parity is Even (set), Odd (cleared)
					U8		SHE : 1;	// bit[4]: Software Handshake Enable (Xon/Xoff)
					U8		HHE : 1;	// bit[5]: Hardware Handshake Enable (RTS/CTS)
					U8		NU3 : 1;	// bit[6]: Not used
					U8		NU4 : 1;	// bit[7]: Not used
				} StatusByte6;
			};

			//Send Message Automatically for Status Register
			union uSMAByte {
				U8 RawSMAB = 0;
				struct sSMAByte {
					// Contains a 1-byte Send Message Automatically status register. The MFC3045 uses the following bits:
					U8	DRST1 : 1;		// Data Ready (Bit 0 of Status Register 1)
					U8	CEST1 : 1;		// Commande Error (Bit 1 of Status Register 1)
					U8	MEST1 : 1;		// Modulation Error (Bit 2 of Status Register 1)
					U8	RSST1 : 1;		// Communication Error (Bit 3  of Status Register 1)
					U8	EEST1 : 1;		// Miscellaneous Error EEPROM (Bit 4  of Status Register 1)
					U8	DNRB : 1;		// Remote Box Button Depressed (down)
					U8	UPRB : 1;		// Remote Box Button Released (Up)
					U8	PANST2 : 1;		// Probe-Array not connected (Bit 7 of Status Register 2)
				} SMAByte;
			};

			//----------------------------------------------------------------------//
			//						Probe Array Information	types					//
			//----------------------------------------------------------------------//
			// odd harmonic of fundamental RF
			enum class eRFH : U8 {
				k1 = 1,
				k3 = 3,
				k5 = 5,
				k7 = 7
			};
			//----------------------------------------------------------------------//
			//					RS232 Communication types	 						//
			//----------------------------------------------------------------------//
			// Serial baudrate
			enum class eMFCSerialBDR : U8 {
				k2400 = 0,
				k4800 = 1,
				k9600 = 2,
				k19200 = 3,
				k28800 = 4,
				k38400 = 5,
				k57600 = 6,
				k115200 = 7
			};

			//RS232 Parameter byte for command "RSP,x"
			union uRSPByte {
				U16 RawRSPB = 0;
				struct sRSPByte {
					// contains a 2-bytes RS232 Parameter. The MFC3045 uses the following bits:
					U8		PBL8 : 1;	// bit[0]: Bit length is 8 (set), 7 (cleared)
					U8		PSB2 : 1;	// bit[1]: Stop bit is 2 (set), 1 (cleared)
					U8		PPAS : 1;	// bit[2]: Parity set
					U8		PPAE : 1;	// bit[3]: Parity is Even (set), Odd (cleared)
					U8		PSHE : 1;	// bit[4]: Software Handshake Enable (Xon/Xoff)
					U8		PHHE : 1;	// bit[5]: Hardware Handshake Enable (RTS/CTS)
					U8		PNU3 : 1;	// bit[6]: Not used
					U8		PNU4 : 1;	// bit[7]: Not used
					U8		PRBR : 3;	// bit[8-10]; RS-232 Baud Rate
										//	0 0 0 2400
										//	0 0 1 4800
										//	0 1 0 9600  (Default value)
										//	0 1 1 19200
										//	1 0 0 28800
										//	1 0 1 38400
										//	1 1 0 57600
										//	1 1 1 115200
					U8		PNU5 : 5;	// bit[11-15]: Not used
				}RSPByte;
			};

			//----------------------------------------------------------------------//
			//						Advance	Command Mode types								//
			//----------------------------------------------------------------------//
			// Advance Command Mode
			enum class eAdvancedMode : U8 {
				kNone = 0,
				k1 = 1,
				k2 = 2
			};

			//----------------------------------------------------------------------//
			//					Devices Version Selection							//
			//----------------------------------------------------------------------//
			enum class eWhoVersion : U8 {
				kMUFirmVer = 0,		// MFC3045 Main unit firmware Version
				kMUEepromVer,		// Main Unit EEPROM Version
				kPAEepromVer		// Probe Array Version
			};

			//----------------------------------------------------------------------//
			//						Devices	Serial Number							//
			//----------------------------------------------------------------------//
			enum class eWhoSerialNum : U8 {
				kMUSerialNum = 0,	// Main Unit Serial Number
				kMUSerialNumBis,	// Main Unit Serial Number
				kPASerialNum		// Probe Array Serial Number
			};

			//----------------------------------------------------------------------//
			//							Remote Box Led								//
			//----------------------------------------------------------------------//
			enum class eRemoteBusyLed : U8 {
				kRemoteLedOff = 0,			// Remote Led is turned OFF
				kRemoteLedMeasuring,		// Remote Led is turned ON when measuring and turned OFF otherwise
				kRemoteLedOffEndMeas,		// Remote Led is turned ON immediately and turned OFF automatically with the end of measurement
				kRemoteLedOn,				// Remote Led is turned ON
				kRemoteLedBlinkSlow,		// Remote Led is blinking slowly (about 1Hz)
				kRemoteLedBlinkFast			// Remote Led is blinking quickly (about 2Hz)
			};

			//----------------------------------------------------------------------//
			//						Frequency Sweep Shift Mode							//
			//----------------------------------------------------------------------//
			enum class eSweepFreqMode : U8 {
				kdisable = 0,
				k20Step,
				kPaRange,
				kSetValue
			};

			//----------------------------------------------------------------------//
			//						Shift Frequency Up/Down							//
			//----------------------------------------------------------------------//
			enum class eShiftFreq : U8 {
				kDown = 2,
				kUp = 8
			};

			//----------------------------------------------------------------------//
			//					Firmware Update Returned State						//
			//----------------------------------------------------------------------//
			enum class eFwUpdateState : U8 {
				kDone = 0,
				kFailed,
				kFatalError,
				kUnkownState
			};

		}	//namespace MFC3045Types
	}	//namesapce Instrument
}	//namespace MTL
