// Standard includes
#include <sstream>
#include <cfloat>

// Personal includes
#include "PT2026TypeConversions.h"
#include "VISAInstrumentBuffer.h"
#include "Helpers.h"
#include "OSDefines.h"
#include "SCPIParsing.h"

using namespace MTL::Instrument;

namespace MTL {
	namespace Instrument {
		namespace PT2026Types {

			U16 BinaryToU16(const char pBinary[2])
			{
				return *reinterpret_cast<const U16*>(pBinary);
			}
			I16 BinaryToI16(const char pBinary[2])
			{
				return *reinterpret_cast<const I16*>(pBinary);
			}
			U32 BinaryToU32(const char pBinary[4])
			{
				return *reinterpret_cast<const U32*>(pBinary);
			}
			I32 BinaryToI32(const char pBinary[4])
			{
				return *reinterpret_cast<const I32*>(pBinary);
			}
			F32 BinaryToF32(const char pBinary[4])
			{
				return *reinterpret_cast<const F32*>(pBinary);
			}
			U64 BinaryToU64(const char pBinary[8])
			{
				return *reinterpret_cast<const U64*>(pBinary);
			}
			F64 BinaryToF64(const char pBinary[8])
			{
				const char Binary64_Out[8] = { pBinary[4], pBinary[5], pBinary[6], pBinary[7], pBinary[0], pBinary[1], pBinary[2], pBinary[3] };
				return *reinterpret_cast<const F64*>(Binary64_Out);
			}
			bool StringToUnits(std::string SUnits, eUnits & rUnits)
			{
				if		(SUnits == "T")		rUnits = kT;
				else if (SUnits == "MT")	rUnits = kmT;
				else if (SUnits == "GAUS")	rUnits = kGauss;
				else if (SUnits == "KGAU")	rUnits = kkGauss;
				else if (SUnits == "PPM")	rUnits = kppm;
				else if (SUnits == "MAHZP")	rUnits = kMHzp;
				else if (SUnits == "MAHZ")	rUnits = kMHz;
				else return false;

				return true;
			}
			std::string UnitsToString(eUnits Units)
			{
				return	Units == kT ?		"T" :
						Units == kmT ?		"MT" :
						Units == kGauss ?	"GAUS" :
						Units == kkGauss ?	"KGAU" :
						Units == kppm ?		"PPM" :
						Units == kMHzp ?	"MAHZP" :
						Units == kMHz ?		"MAHZ" : "";
			}
			bool StringtoExtendedChannel(std::string sExtChan, sExtendedChannel & rExtChan)
			{
				// Separate elements
				std::vector<std::string> l_Tokens;
				SCPI::SplitString(sExtChan, l_Tokens, ',');
				if (l_Tokens.size() != 6)
					return false;
				// Channel
				std::vector<std::string> l_ChannelList;
				SCPI::SplitString(l_Tokens[0], l_ChannelList, '!');
				if (l_ChannelList.empty())
					return false;
				for (auto & ChanIt : l_ChannelList)
					rExtChan.Channel.push_back(static_cast<unsigned char>(std::stoul(ChanIt)));
				// Type
				// TODO: Enable this code for next ARM release
				/*bool l_TypeRecognized = false;
				for (int l_Index = 0; !l_TypeRecognized && l_Index < ePeriphObjectType::NbTypes; l_Index++)
				{
					if (l_Tokens[1] == PeriphObjectTypeStr[l_Index])
					{
						rExtChan.Type = static_cast<ePeriphObjectType>(l_Index);
						l_TypeRecognized = true;
					}
				}
				if (!l_TypeRecognized)
					return false;*/
				// TODO: Remove this code for next ARM release
				rExtChan.Type = (	l_Tokens[1] == "SinglePulsed"		? kPwProbe :
									l_Tokens[1] == "Multiplexer"		? kPwMultiplexer :
									l_Tokens[1] == "SwitchedPulsed"		? kPwCameraHead :
									l_Tokens[1] == "CameraPulsed"		? kPwCamera :
																		  NbTypes);
				if (rExtChan.Type == NbTypes)
					return false;
				// Model
				rExtChan.Model = std::stoul(l_Tokens[2]);
				// Serial
				rExtChan.SerialNumber = std::stoul(l_Tokens[3]);
				// NoPorts
				rExtChan.NoPorts = static_cast<U8>(std::stoul(l_Tokens[4]));
				// CanMeasure
				if (l_Tokens[5] == "1")
					rExtChan.CanMeasure = true;
				else if (l_Tokens[5] == "0")
					rExtChan.CanMeasure = false;
				else
					return false;

				return true;
			}
			bool StringToExtendedChannelList(std::string SExtChanList, CExtendedChannelList & rExtChanList)
			{
				for (size_t l_beg = 0, l_end = 0; l_beg < SExtChanList.size() && l_end < SExtChanList.size(); )
				{
					l_beg = SExtChanList.find_first_of('(', l_end);
					l_end = SExtChanList.find_first_of(')', l_beg);
					if (l_beg != std::string::npos && l_end > l_beg + 2)
					{
						sExtendedChannel l_ExtChan;
						if (StringtoExtendedChannel(SExtChanList.substr(l_beg + 1, l_end - l_beg - 1), l_ExtChan))
							rExtChanList.push_back(l_ExtChan);
						else
							return false;
					}
				}

				return true;
			}

			//----------------------------------------------------------------------//
			//									TESTS								//
			//----------------------------------------------------------------------//
			bool TestPT2026Types()
			{
				try
				{
					//----------------------------------------------------------------------//
					//								Binary Numbers							//
					//----------------------------------------------------------------------//
					unsigned char l_Binary16[2] = { 0x12, 0xF3 };
					if (BinaryToU16(reinterpret_cast<char *>(l_Binary16)) != 62226)
						throw "Invalid U16 conversion";
					if (BinaryToI16(reinterpret_cast<char *>(l_Binary16)) != -3310)
						throw "Invalid I16 conversion";
					unsigned char l_Binary32[4] = { 0x12, 0x34, 0x56, 0xF7 };
					if (BinaryToU32(reinterpret_cast<char *>(l_Binary32)) != 4149621778)
						throw "Invalid U32 conversion";
					if (BinaryToI32(reinterpret_cast<char *>(l_Binary32)) != -145345518)
						throw "Invalid I32 conversion";
					if (BinaryToF32(reinterpret_cast<char *>(l_Binary32)) >= -4.34456E+33F - FLT_EPSILON && BinaryToF32(reinterpret_cast<char *>(l_Binary32)) <= -4.34456E+33F + FLT_EPSILON)
						throw "Invalid F32 conversion";
					unsigned char l_Binary64[8] = { 0x12, 0x34, 0x56, 0x78, 0x9A, 0xBC, 0xDE, 0xF1 };
					if (BinaryToU64(reinterpret_cast<char *>(l_Binary64)) != 17428574979600299026UL)
						throw "Invalid U64 conversion";
					if (BinaryToF64(reinterpret_cast<char *>(l_Binary64)) >= 4.69198E+271 - DBL_EPSILON && BinaryToF64(reinterpret_cast<char *>(l_Binary64)) <= 4.69198E+271 + DBL_EPSILON)
						throw "Invalid F64 conversion";
					CExtendedChannelList l_ExChL;
					std::string l_SExChL("(1,Multiplexer,6026,6,4,0),(1!2,SinglePulsed,1226,57,0,1),(1!4,SinglePulsed,1226,0,0,1)");
					if (!StringToExtendedChannelList(l_SExChL, l_ExChL) || l_ExChL.size() != 3)
						throw "Invalid CExtendedChannelList conversion";
					sExtendedChannel l_ExChan = l_ExChL[1];
					if (l_ExChan.Channel != SCPI::tChannel({ 1,2 }) ||
						l_ExChan.Type != kPwProbe ||
						l_ExChan.Model != 1226 ||
						l_ExChan.SerialNumber != 57 ||
						l_ExChan.NoPorts != 0 ||
						l_ExChan.CanMeasure != true)
						throw "Invalid CExtendedChannelList conversion";
				}
				catch (const char * ErrorMsg)
				{
					COUT("Error! :" << ErrorMsg << std::endl);
					return false;
				}
				COUT(MTL__FUNCTION_NAME__ << " Success" << std::endl);
				return true;
			}

		} // namespace PT2026Types
	} // namespace Instrument
} // namespace MTL
