//////////////////////////////////////////////////////////////////////////
/// \file
/// \brief C++ wrapper for NI-VISA: interface definition.
///
///	For details, please refer to the NI-VISA Programmer Reference Manual
/// (http://www.ni.com/pdf/manuals/370132c.pdf)

#pragma once

// External tools
#include "visatype.h"
#include "visa.h"

// Standard includes
#include <iterator>

// Personal includes
#include "VISAInstrumentTypes.h"
#include "VISAInstrumentBuffer.h"
#include "Synchronization.h"

using namespace MTL::Synchronization;

namespace MTL {
	namespace Instrument {
		//----------------------------------------------------------------------//
		/// \brief Return user-readable description of the given status code.
		/// \param[in]	Session		VISA session.
		/// \param[in]	Status		Status code to be described.
		/// \return String with description of given status code.
		std::string StatusDescription(ViSession Session, ViStatus Status);

		//----------------------------------------------------------------------//
		/// \brief VISA Resource Manager class.
		class CVISAResourceManager
		{
		private:
			mutable CMutex	m_Lock;
			ViSession		m_DefaultRM;
			ViStatus		m_Status;
		public:
			// Constructors / destructors
			CVISAResourceManager();
			virtual ~CVISAResourceManager();

			// Initialization / Info
			bool Initialize();
			ViStatus Status();
			const ViSession & Session();

			// Resource utilities
			bool FindResources(CResourceList & rList, std::string Filter = "?*");
			bool FindResources(CParsedResourceList & rList, std::string Filter = "?*");
			bool ResourceInfo(tResourceName & rRsrc, eInterfaceType & rIntfType, tInterfaceNumber & rIntfNumber);
			bool ResourceInfo(tResourceName & rRsrc, VISAResourceInfo & rReturnedInfo);
		};

		//----------------------------------------------------------------------//
		/// \brief VISA instrument class.
		class CVISAInstrument
		{
		protected:
			mutable CRecursiveMutex		m_Lock;				// Lock onto the class interface
			CVISAResourceManager &		m_rRrsrcMan;		// Reference to the associated resource manager
			tResourceName				m_Rsrc;				// Resource name of the instrument
			ViSession					m_InstrSession;		// Open session
			ViStatus					m_Status;			// Current status

		public:
			// Constructors / destructors
			CVISAInstrument(CVISAResourceManager & rRM, tResourceName Rsrc);
			virtual ~CVISAInstrument();

			// Connexion
			bool Open(eOpenAccessMode AccessMode = eOpenAccessMode::NoLock, ViUInt32 Timeout_ms = 0);
			void Close();
			bool IsOpen();

			// Info
			const ViStatus & Status();
			const ViSession & Session();

			// Write / Read
			bool Write(const char * Str);
			bool Write(const std::string & rStr);
			bool Write(const CVISABuffer & rBuf);
			bool Write(const ViBuf WriteBuf, ViUInt32 BuffLen);
			bool Read(CVISABuffer & rBuf, bool Append = false);
			bool Read(ViPBuf ReadBuf, ViUInt32 ReadLen, ViUInt32 & rRetLen);

			// Other operations
			bool SetTimeout(ViUInt32 Timeout);
			bool Clear();
			bool ReadSTB(ViUInt16 & rSTB);
			bool AssertTrigger(eTriggerProtocol Protocol);

			// Lock / Unlock
			bool LockExclusive(ViUInt32 Timeout);
			bool LockShared(ViUInt32 Timeout, ViKeyId RequestedKey, ViChar AccessKey[]);
			bool Unlock();

			// Events
			bool EnableEvent(eEventType Type, eEventMechanism Mechanism);
			bool DisableEvent(eEventType Type, eEventMechanism Mechanism);
			bool WaitOnEvent(eEventType Type, ViUInt32 Timeout);
			bool WaitOnAllEvents(ViUInt32 Timeout);
			bool DiscardEvents(eEventType Type, eEventMechanism Mechanism);

			// Attributes
			bool GetAttribute(ViAttr Attribute, void * Value);

			// USB
			bool UsbControlIn(ViInt16 bmRequestType, ViInt16 bRequest, ViUInt16 wValue, ViUInt16 wIndex, ViUInt16 wLength, ViPBuf buf, ViUInt16 & rretCnt);
			bool UsbControlOut(ViInt16 bmRequestType, ViInt16 bRequest, ViUInt16 wValue, ViUInt16 wIndex, ViUInt16 wLength, ViBuf buf);

			// Serial
			bool ConfigSerialPort(const sSerialPortSettings & rPortSettings);
		};

}}	// namespace MTL::Instrument
